"""Entry base class"""

import logging

from abc import ABC as AbstractBaseClass, abstractmethod
from typing import Optional

from archey.configuration import Configuration


class Entry(AbstractBaseClass):
    """Module base class"""
    _PRETTY_NAME: Optional[str] = None

    def __new__(cls, *_, **kwargs):
        """Hook object instantiation to handle our particular `disabled` config field"""
        if kwargs.get('options', {}).pop('disabled', False):
            return None

        return super().__new__(cls)

    @abstractmethod
    def __init__(self, name: str = None, value=None, options: dict = None):
        # Each entry will have always have the following attributes...
        # `name`: key (defaults to the instantiated entry class name);
        # `value`: value of entry as an appropriate object;
        # `options`: configuration options *specific* to an entry instance;
        self.name = name or self._PRETTY_NAME or self.__class__.__name__
        self.value = value
        self.options = options or {}

        # Propagates a reference to default strings specified in `Configuration`.
        self._default_strings = Configuration().get('default_strings')

        # Provision a logger for each entry.
        self._logger = logging.getLogger(self.__module__)

    def output(self, output):
        """Output the results to output. Can be overridden by subclasses."""
        if self.value:
            # Let's assume we can just use `__str__` on the object in value,
            # and create a single-line output with it.
            output.append(self.name, str(self.value))
        else:
            # If the value is "falsy" leave a generic "Not detected" message for this entry.
            output.append(
                self.name,
                self._default_strings.get('not_detected')
            )



"""Local IP addresses detection class"""

import ipaddress

from itertools import islice
from typing import Iterator

try:
    import netifaces
except ImportError:
    netifaces = None

from archey.entry import Entry


class LanIP(Entry):
    """Relies on the `netifaces` module to detect LAN IP addresses"""
    _PRETTY_NAME = 'LAN IP'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if not netifaces:
            self._logger.warning(
                "`netifaces` Python module couldn\'t be found. "
                "Please either install it or explicitly disable `LAN_IP` entry in configuration."
            )
            return

        # IPv4 will be enabled by default.
        addr_families = [netifaces.AF_INET]
        if self.options.get('ipv6_support', True):
            addr_families.append(netifaces.AF_INET6)

        max_count = self.options.get('max_count', 2)
        # Consistency with other entries' configuration: Infinite count if false.
        if max_count is False:
            max_count = None

        # Global IP addresses (in RFC1918 terms) will be hidden by default.
        show_global = self.options.get('show_global')

        self.value = list(
            islice(
                self._lan_ip_addresses_generator(addr_families, show_global),
                max_count
            )
        )

    @staticmethod
    def _lan_ip_addresses_generator(addr_families: list, show_global: bool) -> Iterator[str]:
        """Generator yielding local IP address according to passed address families"""
        # Loop through all available network interfaces.
        for if_name in netifaces.interfaces():
            # Fetch associated addresses elements.
            if_addrs = netifaces.ifaddresses(if_name)

            for addr_family in addr_families:
                for if_addr in if_addrs.get(addr_family, []):
                    # IPv6 addresses may contain '%' token separator.
                    ip_addr = ipaddress.ip_address(if_addr['addr'].split('%')[0])

                    # Filter out loopback and public IP addresses.
                    if not ip_addr.is_loopback and (not ip_addr.is_global or show_global):
                        # Finally, yield the address compressed representation.
                        yield ip_addr.compressed


    def output(self, output):
        """Adds the entry to `output` after pretty-formatting the IP address list."""
        # If we found IP addresses, join them together nicely.
        # If not, fall back on default strings according to `netifaces` availability.
        if self.value:
            if not self.options.get('one_line', True):
                # One-line output has been disabled, add one IP address per item.
                for ip_address in self.value:
                    output.append(self.name, ip_address)

                return

            text_output = ', '.join(self.value)

        elif netifaces:
            text_output = self._default_strings.get('no_address')
        else:
            text_output = self._default_strings.get('not_detected')

        output.append(self.name, text_output)



"""Test module for Archey's CPU detection module"""

import unittest
from unittest.mock import MagicMock, call, mock_open, patch

from archey.configuration import DEFAULT_CONFIG
from archey.entries.cpu import CPU
from archey.test import CustomAssertions
from archey.test.entries import HelperMethods


class TestCPUEntry(unittest.TestCase, CustomAssertions):
    """
    Here, we mock the `open` call on `/proc/cpuinfo` with fake content.
    In some cases, `lscpu` output is being mocked too.
    """
    @patch(
        'archey.entries.cpu.open',
        mock_open(
            read_data="""\
processor\t: 0
vendor_id\t: CPU-VENDOR-NAME
cpu family\t: X
model\t\t: YY
model name\t: CPU-MODEL-NAME
physical id\t: 0
"""))
    def test_parse_proc_cpuinfo_one_entry(self):
        """Test `/proc/cpuinfo` parsing"""
        self.assertListEqual(
            CPU._parse_proc_cpuinfo(),  # pylint: disable=protected-access
            [{'CPU-MODEL-NAME': 1}]
        )

    @patch(
        'archey.entries.cpu.open',
        mock_open(
            read_data="""\
processor\t: 0
vendor_id\t: CPU-VENDOR-NAME
cpu family\t: X
model\t\t: YY
model name\t: CPU-MODEL-NAME
physical id\t: 0

processor\t: 1
vendor_id\t: CPU-VENDOR-NAME
cpu family\t: X
model\t\t: YY
model name\t: ANOTHER-CPU-MODEL
physical id\t: 1

processor\t: 2
vendor_id\t: CPU-VENDOR-NAME
cpu family\t: X
model\t\t: YY
model name\t: ANOTHER-CPU-MODEL
physical id\t: 1
"""))
    def test_parse_proc_cpuinfo_multiple_entries(self):
        """Test `/proc/cpuinfo` parsing"""
        self.assertListEqual(
            CPU._parse_proc_cpuinfo(),  # pylint: disable=protected-access
            [
                {'CPU-MODEL-NAME': 1},
                {'ANOTHER-CPU-MODEL': 2}
            ]
        )

    @patch(
        'archey.entries.cpu.open',
        mock_open(
            read_data="""\
processor\t: 0
vendor_id\t: CPU-VENDOR-NAME
cpu family\t: X
model\t\t: YY
model name\t: CPU-MODEL-NAME
physical id\t: 0

processor\t: 1
vendor_id\t: CPU-VENDOR-NAME
cpu family\t: X
model\t\t: YY
model name\t: CPU-MODEL-NAME
physical id\t: 1

processor\t: 2
vendor_id\t: CPU-VENDOR-NAME
cpu family\t: X
model\t\t: YY
model name\t: CPU-MODEL-NAME
physical id\t: 0

processor\t: 3
vendor_id\t: CPU-VENDOR-NAME
cpu family\t: X
model\t\t: YY
model name\t: CPU-MODEL-NAME
physical id\t: 1
"""))
    def test_parse_proc_cpuinfo_one_cpu_dual_socket(self):
        """Test `/proc/cpuinfo` parsing for same CPU model across two sockets"""
        self.assertListEqual(
            CPU._parse_proc_cpuinfo(),  # pylint: disable=protected-access
            [
                {'CPU-MODEL-NAME': 2},
                {'CPU-MODEL-NAME': 2}
            ]
        )

    @patch(
        'archey.entries.cpu.open',
        mock_open(
            read_data="""\
processor\t: 0
vendor_id\t: CPU-VENDOR-NAME
cpu family\t: X
model\t\t: YY
model name\t: CPU-MODEL-NAME
physical id\t: 0

processor\t: 1
vendor_id\t: CPU-VENDOR-NAME
cpu family\t: X
model\t\t: YY
model name\t: ANOTHER-CPU-MODEL
physical id\t: 0

processor\t: 2
vendor_id\t: CPU-VENDOR-NAME
cpu family\t: X
model\t\t: YY
model name\t: ANOTHER\tCPU   MODEL WITH STRANGE S P  A   C     E     S
physical id\t: 1

processor\t: 3
vendor_id\t: CPU-VENDOR-NAME
cpu family\t: X
model\t\t: YY
model name\t: ANOTHER\tCPU   MODEL WITH STRANGE S P  A   C     E     S
physical id\t: 1
"""))
    def test_parse_proc_cpuinfo_multiple_inconsistent_entries(self):
        """
        Test `/proc/cpuinfo` parsing with multiple CPUs sharing same physical ids (unlikely).
        Also check our model name normalizations applied on white characters.
        """
        self.assertListEqual(
            CPU._parse_proc_cpuinfo(),  # pylint: disable=protected-access
            [
                {
                    'CPU-MODEL-NAME': 1,
                    'ANOTHER-CPU-MODEL': 1
                },
                {
                    'ANOTHER CPU MODEL WITH STRANGE S P A C E S': 2
                }
            ]
        )

    @patch(
        'archey.entries.cpu.open',
        side_effect=PermissionError()
    )
    def test_parse_proc_cpuinfo_unreadable_file(self, _):
        """Check behavior when `/proc/cpuinfo` could not be read from disk"""
        self.assertListEmpty(CPU._parse_proc_cpuinfo())  # pylint: disable=protected-access

    @patch(
        'archey.entries.cpu.check_output',
        side_effect=[
            FileNotFoundError(),
            """\
{
  "SPHardwareDataType" : [
    {
      "_name" : "hardware_overview",
      "cpu_type" : "Dual-Core Intel Core i5",
      "current_processor_speed" : "3,1 GHz",
      "machine_model" : "MacBookPro00OOooOO00",
      "machine_name" : "MacBook Pro",
      "number_processors" : 2,
      "packages" : 1,
      "platform_cpu_htt" : "htt_enabled",
      "platform_UUID" : "XXXXXXXX-YYYY-ZZZZ-TTTT-UUUUUUUUUUUU",
      "provisioning_UDID" : "XXXXXXXX-YYYY-ZZZZ-TTTT-UUUUUUUUUUUU",
      "serial_number" : "XX00YY11ZZ22"
    }
  ]
}
"""])
    def test_parse_system_profiler(self, _):
        """Check `_parse_system_profiler` behavior"""
        # pylint: disable=protected-access
        self.assertListEmpty(CPU._parse_system_profiler())
        self.assertListEqual(
            CPU._parse_system_profiler(),
            [{'Dual-Core Intel Core i5 @ 3.1 GHz': 4}]
        )
        # pylint: enable=protected-access

    @patch(
        'archey.entries.cpu.check_output',
        side_effect=[
            FileNotFoundError(),
            """\
Intel(R) Core(TM) i5-7267U CPU @ 3.10GHz
8
"""])
    def test_parse_sysctl_machdep(self, _):
        """Check `_parse_sysctl_machdep` behavior"""
        # pylint: disable=protected-access
        self.assertListEmpty(CPU._parse_sysctl_machdep())
        self.assertListEqual(
            CPU._parse_sysctl_machdep(),
            [{'Intel(R) Core(TM) i5-7267U CPU @ 3.10GHz': 8}]
        )
        # pylint: enable=protected-access

    @patch(
        'archey.entries.cpu.check_output',
        side_effect=[
            """\
Architecture:        x86_64
CPU op-mode(s):      32-bit, 64-bit
Byte Order:          Little Endian
CPU(s):              4
On-line CPU(s) list: 0-3
Thread(s) per core:  1
Core(s) per socket:  4
Socket(s):           1
NUMA node(s):        1
Vendor ID:           CPU-VENDOR-NAME
CPU family:          Z
Model:               \xde\xad\xbe\xef
Model name:          CPU-MODEL-NAME
""",
            """\
Architecture:        x86_64
CPU op-mode(s):      32-bit, 64-bit
Byte Order:          Little Endian
CPU(s):              2
On-line CPU(s) list: 0-1
Thread(s) per core:  1
Core(s) per socket:  2
Socket(s):           1
NUMA node(s):        1
Vendor ID:           CPU-VENDOR-NAME
CPU family:          Z
Model:               \xde\xad\xbe\xef
Model name:          CPU-MODEL-NAME

Architecture:        x86_64
CPU op-mode(s):      32-bit, 64-bit
Byte Order:          Little Endian
CPU(s):              4
On-line CPU(s) list: 0-3
Thread(s) per core:  2
Core(s) per socket:  2
Socket(s):           1
NUMA node(s):        1
Vendor ID:           CPU-VENDOR-NAME
CPU family:          Z
Model:               \xde\xad\xbe\xef
Model name:          ANOTHER-CPU-MODEL
""",
            """\
Architecture:        x86_64
CPU op-mode(s):      32-bit, 64-bit
Byte Order:          Little Endian
CPU(s):              16
On-line CPU(s) list: 0-15
Thread(s) per core:  2
Core(s) per socket:  4
Socket(s):           2
NUMA node(s):        2
Vendor ID:           CPU-VENDOR-NAME
CPU family:          Z
Model:               \xde\xad\xbe\xef
Model name:          CPU-MODEL-NAME
"""])
    def test_parse_lscpu_output(self, _):
        """
        Test model name parsing from `lscpu` output.

        See issue #29 (ARM architectures).
        `/proc/cpuinfo` will not contain `model name` info.
        `lscpu` output will be used instead.
        """
        with self.subTest('Simple unique CPU.'):
            self.assertListEqual(
                CPU._parse_lscpu_output(),  # pylint: disable=protected-access
                [{'CPU-MODEL-NAME': 4}]
            )

        with self.subTest('Two CPUs, 1 socket.'):
            self.assertListEqual(
                CPU._parse_lscpu_output(),  # pylint: disable=protected-access
                [
                    {'CPU-MODEL-NAME': 2},
                    {'ANOTHER-CPU-MODEL': 4}
                ]
            )

        with self.subTest('1 CPU, 2 sockets.'):
            self.assertListEqual(
                CPU._parse_lscpu_output(),  # pylint: disable=protected-access
                [
                    {'CPU-MODEL-NAME': 8},
                    {'CPU-MODEL-NAME': 8}
                ]
            )

    @HelperMethods.patch_clean_configuration
    def test_various_output_configuration(self):
        """Test `output` overloading based on user preferences combination"""
        cpu_instance_mock = HelperMethods.entry_mock(CPU)
        output_mock = MagicMock()

        cpu_instance_mock.value = [
            {'CPU-MODEL-NAME': 1},
            {'ANOTHER-CPU-MODEL': 2}
        ]

        with self.subTest('Single-line combined output.'):
            cpu_instance_mock.options['one_line'] = True

            CPU.output(cpu_instance_mock, output_mock)
            output_mock.append.assert_called_once_with(
                'CPU',
                'CPU-MODEL-NAME, 2 x ANOTHER-CPU-MODEL'
            )

        output_mock.reset_mock()

        with self.subTest('Single-line combined output (no count).'):
            cpu_instance_mock.options['show_cores'] = False
            cpu_instance_mock.options['one_line'] = True

            CPU.output(cpu_instance_mock, output_mock)
            output_mock.append.assert_called_once_with(
                'CPU',
                'CPU-MODEL-NAME, ANOTHER-CPU-MODEL'
            )

        output_mock.reset_mock()

        with self.subTest('Multi-lines output (with counts).'):
            cpu_instance_mock.options['show_cores'] = True
            cpu_instance_mock.options['one_line'] = False

            CPU.output(cpu_instance_mock, output_mock)
            self.assertEqual(output_mock.append.call_count, 2)
            output_mock.append.assert_has_calls(
                [
                    call(
                        'CPU',
                        'CPU-MODEL-NAME'
                    ),
                    call(
                        'CPU',
                        '2 x ANOTHER-CPU-MODEL'
                    )
                ]
            )

        output_mock.reset_mock()

        with self.subTest('No CPU detected output.'):
            cpu_instance_mock.value = []

            CPU.output(cpu_instance_mock, output_mock)
            output_mock.append.assert_called_once_with(
                'CPU',
                DEFAULT_CONFIG['default_strings']['not_detected']
            )


if __name__ == '__main__':
    unittest.main()



"""Test module for Archey's public IP address detection module"""

import unittest
from unittest.mock import MagicMock, Mock, call, patch

from socket import timeout as SocketTimeoutError
from subprocess import TimeoutExpired

from archey.configuration import DEFAULT_CONFIG
from archey.entries.wan_ip import WanIP
from archey.test import CustomAssertions
from archey.test.entries import HelperMethods


@patch(
    'archey.entries.wan_ip.Environment',
    Mock(DO_NOT_TRACK=False)  # By default, entry won't be disabled.
)
class TestWanIPEntry(unittest.TestCase, CustomAssertions):
    """
    Here, we end up mocking calls to `dig` or `urlopen`.
    """
    def setUp(self):
        """We use these mocks so often, it's worth defining them here."""
        self.wan_ip_mock = HelperMethods.entry_mock(WanIP)
        self.output_mock = MagicMock()

    @patch(
        'archey.entries.wan_ip.check_output',
        side_effect=[
            TimeoutExpired('dig', 1),     # `check_output` call will hard-fail.
            '0123::4567:89a:dead:beef\n'  # `check_output` will work.
        ]
    )
    @patch('archey.entries.wan_ip.urlopen')
    def test_ipv4_ko_and_ipv6_ok(self, urlopen_mock, _):
        """Test fallback on HTTP method only when DNS lookup failed"""
        # `urlopen` will hard-fail.
        urlopen_mock.return_value.__enter__.return_value.read.side_effect = SocketTimeoutError(0)

        # IPv4 retrieval failed.
        self.assertFalse(
            WanIP._retrieve_ip_address(self.wan_ip_mock, 4),  # pylint: disable=protected-access
        )

        # IPv6 worked like a (almost !) charm.
        self.assertEqual(
            WanIP._retrieve_ip_address(self.wan_ip_mock, 6),  # pylint: disable=protected-access
            '0123::4567:89a:dead:beef'
        )

    @patch(
        'archey.entries.wan_ip.check_output',
        side_effect=[
            '\n',                     # `check_output` call will soft-fail.
            FileNotFoundError('dig')  # `check_output` call will hard-fail.
        ]
    )
    @patch('archey.entries.wan_ip.urlopen')
    def test_proper_http_fallback(self, urlopen_mock, _):
        """Test fallback on HTTP method only when DNS lookup failed"""
        urlopen_mock.return_value.__enter__.return_value.read.return_value = b'XXX.YY.ZZ.TTT\n'

        # HTTP back-end was not called, we trust DNS lookup tool which failed.
        self.assertFalse(
            WanIP._retrieve_ip_address(self.wan_ip_mock, 4),  # pylint: disable=protected-access
        )

        # New try: HTTP method has been called !
        self.assertEqual(
            WanIP._retrieve_ip_address(self.wan_ip_mock, 4),  # pylint: disable=protected-access
            'XXX.YY.ZZ.TTT'
        )

    def test_retrieval_disabled(self):
        """Test behavior when both IPv4 and IPv6 retrievals are purposely disabled"""
        self.wan_ip_mock.options = {
            'ipv4': False,
            'ipv6': False
        }

        # Both retrievals fail.
        self.assertFalse(
            WanIP._retrieve_ip_address(self.wan_ip_mock, 4)  # pylint: disable=protected-access
        )
        self.assertFalse(
            WanIP._retrieve_ip_address(self.wan_ip_mock, 6)  # pylint: disable=protected-access
        )

    def test_method_disabled(self):
        """Check whether user could disable resolver back-ends from configuration"""
        self.wan_ip_mock.options = {
            'ipv4': {
                'dns_query': False,
                'http_url': False
            }
        }

        # Internal method doesn't return any address.
        self.assertFalse(
            WanIP._retrieve_ip_address(self.wan_ip_mock, 4)  # pylint: disable=protected-access
        )

    def test_two_addresses(self):
        """
        Test when both IPv4 and IPv6 addresses could be retrieved.
        Additionally check the `output` method behavior.
        """
        self.wan_ip_mock.value = ['XXX.YY.ZZ.TTT', '0123::4567:89a:dead:beef']

        with self.subTest('Single-line combined output.'):
            WanIP.output(self.wan_ip_mock, self.output_mock)

            self.assertEqual(
                self.output_mock.append.call_args[0][1],
                "XXX.YY.ZZ.TTT, 0123::4567:89a:dead:beef"
            )

        self.output_mock.reset_mock()

        with self.subTest('Multi-lines output.'):
            self.wan_ip_mock.options['one_line'] = False

            WanIP.output(self.wan_ip_mock, self.output_mock)
            self.assertEqual(self.output_mock.append.call_count, 2)
            self.output_mock.append.assert_has_calls(
                [
                    call('WAN IP', 'XXX.YY.ZZ.TTT'),
                    call('WAN IP', '0123::4567:89a:dead:beef')
                ]
            )

    def test_do_not_track(self):
        """Check whether `DO_NOT_TRACK` environment variable is correctly honored"""
        with patch('archey.entries.wan_ip.Environment', Mock(DO_NOT_TRACK=True)):
            self.wan_ip_mock.value = []

            WanIP.output(self.wan_ip_mock, self.output_mock)

            self.assertListEmpty(self.wan_ip_mock.value)
            self.assertEqual(
                self.output_mock.append.call_args[0][1],
                DEFAULT_CONFIG['default_strings']['not_detected']
            )

    @HelperMethods.patch_clean_configuration
    def test_no_address(self):
        """
        Test when no address could be retrieved.
        Additionally check the `output` method behavior.
        """
        self.wan_ip_mock.value = []

        WanIP.output(self.wan_ip_mock, self.output_mock)

        self.assertListEmpty(self.wan_ip_mock.value)
        self.assertEqual(
            self.output_mock.append.call_args[0][1],
            DEFAULT_CONFIG['default_strings']['no_address']
        )


if __name__ == '__main__':
    unittest.main()



"""Gentoo logo"""

from archey.colors import Colors


COLORS = [Colors.MAGENTA_BRIGHT, Colors.WHITE_BRIGHT]

LOGO = [
    """{c[0]}          -/oyddmdhs+:.             """,
    """{c[0]}      -o{c[1]}dNMMMMMMMMNNmhy+{c[0]}-`          """,
    """{c[0]}    -y{c[1]}NMMMMMMMMMMMNNNmmdhy{c[0]}+-        """,
    """{c[0]}  `o{c[1]}mMMMMMMMMMMMMNmdmmmmddhhy{c[0]}/`     """,
    """{c[0]}  om{c[1]}MMMMMMMMMMMN{c[0]}hhyyyo{c[1]}hmdddhhhd{c[0]}o`   """,
    """{c[0]} .y{c[1]}dMMMMMMMMMMd{c[0]}hs++so/s{c[1]}mdddhhhhdm{c[0]}+` """,
    """{c[0]}  oy{c[1]}hdmNMMMMMMMN{c[0]}dyooy{c[1]}dmddddhhhhyhN{c[0]}d.""",
    """{c[0]}   :o{c[1]}yhhdNNMMMMMMMNNNmmdddhhhhhyym{c[0]}Mh""",
    """{c[0]}     .:{c[1]}+sydNMMMMMNNNmmmdddhhhhhhmM{c[0]}my""",
    """{c[0]}        /m{c[1]}MMMMMMNNNmmmdddhhhhhmMNh{c[0]}s:""",
    """{c[0]}     `o{c[1]}NMMMMMMMNNNmmmddddhhdmMNhs{c[0]}+` """,
    """{c[0]}   `s{c[1]}NMMMMMMMMNNNmmmdddddmNMmhs{c[0]}/.   """,
    """{c[0]}  /N{c[1]}MMMMMMMMNNNNmmmdddmNMNdso{c[0]}:`     """,
    """{c[0]} +M{c[1]}MMMMMMNNNNNmmmmdmNMNdso{c[0]}/-        """,
    """{c[0]} yM{c[1]}MNNNNNNNmmmmmNNMmhs+/{c[0]}-`          """,
    """{c[0]} /h{c[1]}MMNNNNNNNNMNdhs++/{c[0]}-`             """,
    """{c[0]} `/{c[1]}ohdmmddhys+++/:{c[0]}.`                """,
    """{c[0]} `-//////:--.                       """
]



"""Archey Utility module"""

from typing import Tuple

from archey.singleton import Singleton


class Utility(metaclass=Singleton):
    """Miscellaneous logic used in Archey internals"""
    @classmethod
    def update_recursive(cls, old_dict: dict, new_dict: dict):
        """
        A method for recursively merging dictionaries as `dict.update()` is not able to do this.
        Original snippet taken from here : <https://gist.github.com/angstwad/bf22d1822c38a92ec0a9>
        """
        for key, value in new_dict.items():
            if key in old_dict \
                and isinstance(old_dict[key], dict) \
                and isinstance(value, dict):
                cls.update_recursive(old_dict[key], value)
            else:
                old_dict[key] = value

    @staticmethod
    def version_to_semver_segments(version: str) -> Tuple[int, ...]:
        """Transforms string `version` to a tuple containing SemVer segments"""
        return tuple(
            map(
                int,
                version.lstrip('v').partition('-')[0].split('.')
            )
        )



"""Kernel information detection class"""

import json
import platform

from socket import timeout as SocketTimeoutError
from urllib.error import URLError
from urllib.request import urlopen
from typing import Optional

from archey.entry import Entry
from archey.environment import Environment
from archey.utility import Utility


class Kernel(Entry):
    """
    Retrieve kernel identity.
    [GNU/LINUX] If user-enabled, implement a version comparison against upstream data.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.value = {
            'name': platform.system(),
            'release': platform.release(),
            'latest': None,
            'is_outdated': None
        }

        # On GNU/Linux systems, if `check_version` has been enabled and `DO_NOT_TRACK` isn't set,
        #  retrieve the latest kernel release in order to compare the current one against it.
        if not self.options.get('check_version') \
            or self.value['name'] != 'Linux' \
            or Environment.DO_NOT_TRACK:
            return

        self.value['latest'] = self._fetch_latest_linux_release()
        if self.value['latest']:
            self.value['is_outdated'] = (
                Utility.version_to_semver_segments(self.value['release'])
                < Utility.version_to_semver_segments(self.value['latest'])
            )

    @staticmethod
    def _fetch_latest_linux_release() -> Optional[str]:
        try:
            with urlopen('https://www.kernel.org/releases.json') as http_request:
                try:
                    kernel_releases = json.load(http_request)
                except json.JSONDecodeError:
                    return None
        except (URLError, SocketTimeoutError):
            return None

        return kernel_releases.get('latest_stable', {}).get('version')

    def output(self, output):
        """Display running kernel and latest kernel if possible"""
        text_output = ' '.join((self.value['name'], self.value['release']))

        if self.value['latest']:
            if self.value['is_outdated']:
                text_output += f" ({self.value['latest']} {self._default_strings.get('available')})"
            else:
                text_output += f" ({self._default_strings.get('latest')})"

        output.append(self.name, text_output)



"""GPU information detection class"""

import platform
import re

from subprocess import CalledProcessError, DEVNULL, check_output
from typing import List

from archey.entry import Entry


class GPU(Entry):
    """Relies on `lspci` or `pciconf` to retrieve graphical device(s) information"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if platform.system() == 'Linux':
            self.value = self._parse_lspci_output()
        else:
            # Darwin or any other BSD-based system.
            self.value = self._parse_system_profiler() or \
                self._parse_pciconf_output()

        max_count = self.options.get('max_count', 2)
        # Consistency with other entries' configuration: Infinite count if false.
        if max_count is not False:
            self.value = self.value[:max_count]

    @staticmethod
    def _parse_lspci_output() -> List[str]:
        """Based on `lspci` output, return a list of video controllers names"""
        try:
            lspci_output = check_output('lspci', universal_newlines=True).splitlines()
        except (FileNotFoundError, CalledProcessError):
            return []

        gpus_list = []

        # We'll be looking for specific video controllers (in the below keys order).
        for video_key in ('3D', 'VGA', 'Display'):
            for pci_device in lspci_output:
                # If a controller type match...
                if video_key in pci_device:
                    # ... adds its name to the final list.
                    gpus_list.append(pci_device.partition(': ')[2])

        return gpus_list

    @staticmethod
    def _parse_system_profiler() -> List[str]:
        """Based on `system_profiler` output, return a list of video chipsets model names"""
        # Parse output from Darwin's `system_profiler` binary.
        # We do not use JSON output (more than 10 times longer for nothing).
        try:
            profiler_output = check_output(
                ['system_profiler', 'SPDisplaysDataType'],
                stderr=DEVNULL, universal_newlines=True
            )
        except FileNotFoundError:
            return []

        return re.findall(
            r'Chipset Model: (.*)',
            profiler_output,
            re.MULTILINE
        )

    @staticmethod
    def _parse_pciconf_output() -> List[str]:
        """Based on `pciconf` output, return a list of video devices as long as their vendor"""
        try:
            pciconf_output = check_output(
                ['pciconf', '-lv'],
                stderr=DEVNULL, universal_newlines=True
            )
        except (FileNotFoundError, CalledProcessError):
            return []

        gpus_list = []

        for vendor, device in re.findall(
                r'vendor\s*=\s*\'(.*)\'\s*device\s*=\s*\'(.*)\'\s*class\s*=\s*display',
                pciconf_output,
                re.MULTILINE
            ):
            gpus_list.append(f'{vendor} {device}')

        return gpus_list


    def output(self, output):
        """Writes GPUs to `output` based on preferences"""
        # No GPU could be detected.
        if not self.value:
            output.append(self.name, self._default_strings.get('not_detected'))
            return

        # Join the results only if `one_line` option is enabled.
        if self.options.get('one_line'):
            output.append(self.name, ', '.join(self.value))
        else:
            for gpu_device in self.value:
                output.append(self.name, gpu_device)



"""Test module for Archey's RAM usage detection module"""

import unittest
from unittest.mock import mock_open, patch, MagicMock

from archey.colors import Colors
from archey.configuration import DEFAULT_CONFIG
from archey.entries.ram import RAM
from archey.test.entries import HelperMethods


class TestRAMEntry(unittest.TestCase):
    """
    Here, we mock the `check_output` call to `free` using all three levels of available ram.
    In the last test, mock with `/proc/meminfo` file opening during the manual way.
    """
    @patch(
        'archey.entries.ram.check_output',
        return_value="""\
              total        used        free      shared  buff/cache   available
Mem:          15658        2043       10232          12        3382       13268
Swap:          4095          39        4056
""")
    def test_run_free_dash_m(self, _):
        """Test `_run_free_dash_m` output parsing"""
        self.assertTupleEqual(
            RAM._run_free_dash_m(),  # pylint: disable=protected-access
            (2043.0, 15658.0)
        )

    @patch(
        'archey.entries.ram.open',
        mock_open(
            read_data="""\
MemTotal:        7581000 kB
MemFree:          716668 kB
MemAvailable:    3632244 kB
Buffers:          478524 kB
Cached:          2807032 kB
SwapCached:        67092 kB
Active:          3947284 kB
Inactive:        2447708 kB
Active(anon):    2268724 kB
Inactive(anon):  1106220 kB
Active(file):    1678560 kB
Inactive(file):  1341488 kB
Unevictable:         128 kB
Mlocked:             128 kB
SwapTotal:       7811068 kB
SwapFree:        7277708 kB
Dirty:               144 kB
Writeback:             0 kB
AnonPages:       3067204 kB
Mapped:           852272 kB
Shmem:            451056 kB
Slab:             314100 kB
SReclaimable:     200792 kB
SUnreclaim:       113308 kB
"""))  # Some lines have been ignored as they are useless for computations.
    def test_read_proc_meminfo(self):
        """Test `_read_proc_meminfo` content parsing"""
        self.assertTupleEqual(
            RAM._read_proc_meminfo(),  # pylint: disable=protected-access
            (3739.296875, 7403.3203125)
        )

    @patch(
        'archey.entries.ram.check_output',
        side_effect=[
            '8589934592\n',
            """\
Mach Virtual Memory Statistics: (page size of 4096 bytes)
Pages free:                               55114.
Pages active:                            511198.
Pages inactive:                          488363.
Pages speculative:                        22646.
Pages throttled:                              0.
Pages wired down:                        666080.
Pages purgeable:                          56530.
"Translation faults":                 170435998.
Pages copy-on-write:                    3496023.
Pages zero filled:                     96454484.
Pages reactivated:                     12101726.
Pages purged:                           6728288.
File-backed pages:                       445114.
Anonymous pages:                         577093.
Pages stored in compressor:             2019211.
Pages occupied by compressor:            353431.
Decompressions:                        10535599.
Compressions:                          19723567.
Pageins:                                7586286.
Pageouts:                                388644.
Swapins:                                2879182.
Swapouts:                               3456015.
"""])
    def test_run_sysctl_and_vmstat(self, _):
        """Check `sysctl` and `vm_stat` parsing logic"""
        self.assertTupleEqual(
            RAM._run_sysctl_and_vmstat(),  # pylint: disable=protected-access
            (1685.58984375, 8192.0)
        )

    @HelperMethods.patch_clean_configuration
    def test_various_output_configuration(self):
        """Test `output` overloading based on user preferences"""
        ram_instance_mock = HelperMethods.entry_mock(RAM)
        output_mock = MagicMock()

        with self.subTest('Output in case of non-detection.'):
            RAM.output(ram_instance_mock, output_mock)
            self.assertEqual(
                output_mock.append.call_args[0][1],
                DEFAULT_CONFIG['default_strings']['not_detected']
            )

        output_mock.reset_mock()

        with self.subTest('"Normal" output (green).'):
            ram_instance_mock.value = {
                'used': 2043.0,
                'total': 15658.0,
                'unit': 'MiB'
            }
            ram_instance_mock.options = {
                'warning_use_percent': 33.3,
                'danger_use_percent': 66.7
            }

            RAM.output(ram_instance_mock, output_mock)
            self.assertEqual(
                output_mock.append.call_args[0][1],
                f'{Colors.GREEN_NORMAL}2043 MiB{Colors.CLEAR} / 15658 MiB'
            )

        output_mock.reset_mock()

        with self.subTest('"Danger" output (red).'):
            ram_instance_mock.value = {
                'used': 7830.0,
                'total': 15658.0,
                'unit': 'MiB'
            }
            ram_instance_mock.options = {
                'warning_use_percent': 25,
                'danger_use_percent': 50
            }

            RAM.output(ram_instance_mock, output_mock)
            self.assertEqual(
                output_mock.append.call_args[0][1],
                f'{Colors.RED_NORMAL}7830 MiB{Colors.CLEAR} / 15658 MiB'
            )


if __name__ == '__main__':
    unittest.main()



"""Test module for Archey's disks usage detection module"""

import unittest
from unittest.mock import call, patch, MagicMock

from archey.colors import Colors
from archey.entries.disk import Disk
from archey.test.entries import HelperMethods


class TestDiskEntry(unittest.TestCase):
    """
    Here, we mock `subprocess.run` calls to disk utility tools.
    """
    def setUp(self):
        """We use these mocks so often, it's worth defining them here."""
        self.disk_instance_mock = HelperMethods.entry_mock(Disk)
        self.output_mock = MagicMock()

    def test_disk_get_local_filesystems(self):
        """Tests `Disk._get_local_filesystems`."""
        # This minimal `_disk_dict` contains everything this method touches.
        self.disk_instance_mock._disk_dict = {  # pylint: disable=protected-access
            '/very/good/mountpoint': {
                'device_path': '/dev/sda1'
            },
            '/mounted/here/too': {
                'device_path': '/dev/sda1'
            },
            '/other/acceptable/device/paths': {
                'device_path': '/dev/anything-really'
            },
            '/a/samba/share': {
                'device_path': '//server.local/cool_share'  # ignored - not `/dev/...`
            },
            '/linux/loop/device/one': {
                'device_path': '/dev/loop0'  # ignored - loop device
            },
            '/linux/loop/device/two': {
                'device_path': '/dev/blah/loop0'  # ignored - loop device
            },
            '/bsd/s/loop/device/one': {
                'device_path': '/dev/svnd'  # ignored - loop device
            },
            '/bsd/s/loop/device/two': {
                'device_path': '/dev/blah/svnd1'  # ignored - loop device
            },
            '/bsd/r/loop/device/one': {
                'device_path': '/dev/rvnd'  # ignored - loop device
            },
            '/bsd/r/loop/device/two': {
                'device_path': '/dev/blah/rvnd1'  # ignored - loop device
            },
            '/solaris/loop/device/one': {
                'device_path': '/dev/lofi1'  # ignored - loop device
            },
            '/solaris/loop/device/two': {
                'device_path': '/dev/blah/lofi'  # ignored - loop device
            },
            '/linux/device/mapper': {
                'device_path': '/dev/dm-1'  # ignored - device mapper
            }
        }

        self.assertDictEqual(
            Disk._get_local_filesystems(self.disk_instance_mock),  # pylint: disable=protected-access
            {
                '/very/good/mountpoint': {
                    'device_path': '/dev/sda1'
                },
                '/other/acceptable/device/paths': {
                    'device_path': '/dev/anything-really'
                }
            }
        )

    def test_disk_get_specified_filesystems(self):
        """Tests `Disk._get_specified_filesystems`."""
        # This minimal `_disk_dict` contains everything this method touches.
        self.disk_instance_mock._disk_dict = {  # pylint: disable=protected-access
            '/very/good/mountpoint': {
                'device_path': '/dev/sda1'
            },
            '/mounted/here/too': {
                'device_path': '/dev/sda1'
            },
            '/less/good/mountpoint': {
                'device_path': '/dev/sda2'
            },
            '/a/samba/share': {
                'device_path': '//server.local/cool_share'
            }
        }

        with self.subTest('Get all filesystems with mount points.'):
            # pylint: disable=protected-access
            self.assertDictEqual(
                Disk._get_specified_filesystems(
                    self.disk_instance_mock,
                    self.disk_instance_mock._disk_dict  # recall dicts are iterables of their keys.
                ),
                self.disk_instance_mock._disk_dict
            )
            # pylint: enable=protected-access

        with self.subTest('Get only `/dev/sda1` filesystems.'):
            self.assertDictEqual(
                Disk._get_specified_filesystems(  # pylint: disable=protected-access
                    self.disk_instance_mock,
                    ('/dev/sda1',)
                ),
                {
                    '/very/good/mountpoint': {
                        'device_path': '/dev/sda1'
                    }
                }
            )


    @patch('archey.entries.disk.run')
    def test_disk_df_output_dict(self, run_mock):
        """Test method to get `df` output as a dict by mocking calls to `subprocess.run`"""
        with self.subTest('`df` regular output.'):
            run_mock.return_value.stdout = '\n'.join([
                "Filesystem               1024-blocks      Used     Available Capacity Mounted on",
                "/dev/nvme0n1p2             499581952 427458276      67779164      87% /",
                "tmpfs                        8127236       292       8126944       1% /tmp",
                "/dev/nvme0n1p1                523248     35908        487340       7% /boot",
                "/dev/sda1                       1624        42          1582       1% /what is  this",             # pylint: disable=line-too-long
                "map auto_home                      0         0             0     100% /System/Volumes/Data/home",  # pylint: disable=line-too-long
                "/Applications/Among Us.app/Wrapper 0         0             0     100% /System/Volumes/Data/game",  # pylint: disable=line-too-long
                ""
            ])
            self.assertDictEqual(
                Disk._get_df_output_dict(),  # pylint: disable=protected-access
                {
                    '/': {
                        'device_path': '/dev/nvme0n1p2',
                        'used_blocks': 427458276,
                        'total_blocks': 499581952
                    },
                    '/tmp': {
                        'device_path': 'tmpfs',
                        'used_blocks': 292,
                        'total_blocks': 8127236
                    },
                    '/boot': {
                        'device_path': '/dev/nvme0n1p1',
                        'used_blocks': 35908,
                        'total_blocks': 523248
                    },
                    '/what is  this': {
                        'device_path': '/dev/sda1',
                        'used_blocks': 42,
                        'total_blocks': 1624
                    }
                }
            )

        with self.subTest('`df` missing from system.'):
            run_mock.side_effect = FileNotFoundError()
            self.assertDictEqual(
                Disk._get_df_output_dict(),  # pylint: disable=protected-access
                {}
            )

    def test_disk_blocks_to_human_readable(self):
        """Test method to convert 1024-byte blocks to a human readable format."""
        # Each tuple is a number of blocks followed by the expected output.
        test_cases = (
            (1, '1.0 KiB'),
            (1024, '1.0 MiB'),
            (2048, '2.0 MiB'),
            (95604, '93.4 MiB'),
            (1048576, '1.0 GiB'),
            (2097152, '2.0 GiB'),
            (92156042, '87.9 GiB'),
            (1073742000, '1.0 TiB'),
            (2147484000, '2.0 TiB'),
            (458028916298, '426.6 TiB'),
            (1099512000000, '1.0 PiB'),
            (2199023000000, '2.0 PiB')  # I think we can safely stop here :)
        )
        for test_case in test_cases:
            with self.subTest(test_case[1]):
                self.assertEqual(
                    Disk._blocks_to_human_readable(test_case[0]),  # pylint: disable=protected-access
                    test_case[1]
                )

    def test_disk_output_colors(self):
        """Test `output` disk level coloring."""
        # This dict's values are tuples of used blocks, and the level's corresponding color.
        # For reference, this test uses a disk whose total block count is 100.
        levels = {
            'normal': (45.0, Colors.GREEN_NORMAL),
            'warning': (70.0, Colors.YELLOW_NORMAL),
            'danger': (95.0, Colors.RED_NORMAL)
        }
        for level, blocks_color_tuple in levels.items():
            with self.subTest(level):
                self.disk_instance_mock.value = {
                    'mount_point': {
                        'device_path': '/dev/my-cool-disk',
                        'used_blocks': blocks_color_tuple[0],
                        'total_blocks': 100
                    }
                }
                Disk.output(self.disk_instance_mock, self.output_mock)
                self.output_mock.append.assert_called_with(
                    'Disk',
                    f'{blocks_color_tuple[1]}{blocks_color_tuple[0]} KiB{Colors.CLEAR} / 100.0 KiB'
                )

    def test_disk_multiline_output(self):
        """Test `output`'s multi-line capability."""
        self.disk_instance_mock.value = {
            'first_mount_point': {
                'device_path': '/dev/my-cool-disk',
                'used_blocks': 10,
                'total_blocks': 10
            },
            'second_mount_point': {
                'device_path': '/dev/my-cooler-disk',
                'used_blocks': 10,
                'total_blocks': 30
            }
        }

        with self.subTest('Single-line combined output.'):
            Disk.output(self.disk_instance_mock, self.output_mock)
            self.output_mock.append.assert_called_once_with(
                'Disk',
                f'{Colors.YELLOW_NORMAL}20.0 KiB{Colors.CLEAR} / 40.0 KiB'
            )

        self.output_mock.reset_mock()

        with self.subTest('Multi-line output'):
            self.disk_instance_mock.options['combine_total'] = False
            Disk.output(self.disk_instance_mock, self.output_mock)
            self.assertEqual(self.output_mock.append.call_count, 2)
            self.output_mock.append.assert_has_calls(
                [
                    call(
                        'Disk',
                        f'{Colors.RED_NORMAL}10.0 KiB{Colors.CLEAR} / 10.0 KiB'
                    ),
                    call(
                        'Disk',
                        f'{Colors.GREEN_NORMAL}10.0 KiB{Colors.CLEAR} / 30.0 KiB'
                    )
                ]
            )

        self.output_mock.reset_mock()

        with self.subTest('Entry name labeling (device path with entry name)'):
            self.disk_instance_mock.options = {
                'combine_total': False,
                'disk_labels': 'device_paths'
            }

            Disk.output(self.disk_instance_mock, self.output_mock)
            self.assertEqual(self.output_mock.append.call_count, 2)
            self.output_mock.append.assert_has_calls(
                [
                    call(
                        'Disk (/dev/my-cool-disk)',
                        f'{Colors.RED_NORMAL}10.0 KiB{Colors.CLEAR} / 10.0 KiB'
                    ),
                    call(
                        'Disk (/dev/my-cooler-disk)',
                        f'{Colors.GREEN_NORMAL}10.0 KiB{Colors.CLEAR} / 30.0 KiB'
                    )
                ]
            )

        self.output_mock.reset_mock()

        with self.subTest('Entry name labeling (mount points without entry name)'):
            self.disk_instance_mock.options = {
                'combine_total': False,
                'disk_labels': 'mount_points',
                'hide_entry_name': True
            }

            Disk.output(self.disk_instance_mock, self.output_mock)
            self.assertEqual(self.output_mock.append.call_count, 2)
            self.output_mock.append.assert_has_calls(
                [
                    call(
                        '(first_mount_point)',
                        f'{Colors.RED_NORMAL}10.0 KiB{Colors.CLEAR} / 10.0 KiB'
                    ),
                    call(
                        '(second_mount_point)',
                        f'{Colors.GREEN_NORMAL}10.0 KiB{Colors.CLEAR} / 30.0 KiB'
                    )
                ]
            )

        self.output_mock.reset_mock()

        with self.subTest('Entry name labeling (without disk label nor entry name)'):
            self.disk_instance_mock.options = {
                'combine_total': False,
                'disk_labels': False,
                # `hide_entry_name` is being ignored as `disk_labels` evaluates to "falsy" too.
                'hide_entry_name': True
            }

            Disk.output(self.disk_instance_mock, self.output_mock)
            self.assertEqual(self.output_mock.append.call_count, 2)
            self.output_mock.append.assert_has_calls(
                [
                    call(
                        'Disk',
                        f'{Colors.RED_NORMAL}10.0 KiB{Colors.CLEAR} / 10.0 KiB'
                    ),
                    call(
                        'Disk',
                        f'{Colors.GREEN_NORMAL}10.0 KiB{Colors.CLEAR} / 30.0 KiB'
                    )
                ]
            )


if __name__ == '__main__':
    unittest.main()



"""Shell detection class"""

import os

from subprocess import CalledProcessError, check_output

from archey.entry import Entry


class Shell(Entry):
    """
    Simple shell path detection based either on the `SHELL` environment variable or
    the local administrative database.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.value = os.getenv("SHELL") or self._query_name_service_switch()

    @staticmethod
    def _query_name_service_switch():
        try:
            user_id = os.getuid()
        except AttributeError:
            # Not UNIX...
            return None

        try:
            shell = check_output(
                ["getent", "passwd", str(user_id)],
                universal_newlines=True
            ).rstrip().rsplit(":", maxsplit=1)[-1]
        except CalledProcessError:
            # Ghost user...
            return None

        return shell



"""Hardware model information detection class"""

import os
import platform
import re

from subprocess import CalledProcessError, DEVNULL, check_output
from typing import Optional

from archey.entry import Entry


class Model(Entry):
    """Uses multiple methods to retrieve some information about the host hardware"""

    LINUX_DMI_SYS_PATH = "/sys/devices/virtual/dmi/id"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.value = \
            self._fetch_virtual_env_info() \
            or self._fetch_dmi_info() \
            or self._fetch_sysctl_hw() \
            or self._fetch_raspberry_pi_revision() \
            or self._fetch_android_device_model()

    def _fetch_virtual_env_info(self) -> Optional[str]:
        """
        Relying on some system tools, tries to gather some details about hypervisor.
        When available, relies on systemd.
        If running with enough privileges, `virt-what` and/or `dmidecode` may be called.
        """
        try:
            environment = check_output(
                'systemd-detect-virt',
                stderr=DEVNULL, universal_newlines=True
            ).rstrip()
        except CalledProcessError:
            # Not a virtual environment.
            environment = ""
        except FileNotFoundError:
            # If not available, let's query `virt-what` (privileges usually required).
            try:
                environment = ", ".join(
                    check_output(
                        "virt-what",
                        stderr=DEVNULL, universal_newlines=True
                    ).splitlines()
                )
            except (OSError, CalledProcessError):
                environment = ""

        # We couldn't retrieve any information from `systemd-detect-virt`, nor `virt-what`.
        if not environment:
            return None

        # Sometimes we may gather info added by hosting service provider by relying on `dmidecode`.
        try:
            product_name = check_output(
                ["dmidecode", "-s", "system-product-name"],
                stderr=DEVNULL, universal_newlines=True
            ).rstrip()
        except (OSError, CalledProcessError):
            product_name = self._default_strings.get("virtual_environment")

        # If we got there with some info, this _should_ be a virtual environment.
        return f"{product_name} ({environment})"

    @classmethod
    def _fetch_dmi_info(cls) -> Optional[str]:
        """Tries to open DMI Linux files, looking for hardware information"""
        def _read_dmi_file(file_name: str) -> str:
            try:
                with open(
                    os.path.join(cls.LINUX_DMI_SYS_PATH, file_name), encoding='UTF-8'
                ) as f_dmi_file:
                    dmi_info = f_dmi_file.read().rstrip()
            except OSError:
                return ""

            # Stop `/sys/devices/virtual/dmi/id/*` parsing on fuzzy data.
            if "to be filled" in dmi_info.lower():
                return ""

            return dmi_info

        # Fetch product name.
        product_name = _read_dmi_file("product_name")
        if product_name:
            product_info = [product_name]
            # Prepend product vendor name (if available).
            product_info.insert(0, _read_dmi_file("sys_vendor"))
            # Append product version (if available).
            product_info.append(_read_dmi_file("product_version"))

            return " ".join(filter(None, product_info))

        # Fetch board name.
        board_name = _read_dmi_file("board_name")
        if board_name:
            board_info = [board_name]
            # Prepend board vendor name (if available).
            board_info.insert(0, _read_dmi_file("board_vendor"))
            # Append board version (if available).
            board_info.append(_read_dmi_file("board_version"))

            return " ".join(filter(None, board_info))

        return None

    @staticmethod
    def _fetch_sysctl_hw() -> Optional[str]:
        # `hw.model` might be populated with CPU info on BSD platforms.
        # Let's only query this OID on Darwin (macOS).
        if platform.system() == 'Darwin':
            try:
                model = check_output(
                    ['sysctl', '-n', 'hw.model'],
                    stderr=DEVNULL, universal_newlines=True
                )
            except FileNotFoundError:
                return None
            except CalledProcessError:
                pass
            else:
                return model.rstrip().replace(',', '.')

        # Any other BSD (or derivatives).
        hw_oids = []
        for hw_oid in ('vendor', 'product', 'version'):
            try:
                sysctl_output = check_output(
                    ['sysctl', '-n', f'hw.{hw_oid}'],
                    stderr=DEVNULL, universal_newlines=True
                )
            except FileNotFoundError:
                return None
            except CalledProcessError:
                pass
            else:
                sysctl_output = sysctl_output.rstrip()
                if sysctl_output != 'None':
                    hw_oids.append(sysctl_output)

        return ' '.join(hw_oids) or None

    @staticmethod
    def _fetch_raspberry_pi_revision() -> Optional[str]:
        """Tries to retrieve 'Hardware' and 'Revision IDs' from `/proc/cpuinfo`"""
        try:
            with open('/proc/cpuinfo', encoding='ASCII') as f_cpu_info:
                cpu_info = f_cpu_info.read()
        except OSError:
            return None

        # If the output contains 'Hardware' and 'Revision'...
        hardware = re.search(r'(?<=Hardware\t: ).*', cpu_info)
        revision = re.search(r'(?<=Revision\t: ).*', cpu_info)
        if hardware and revision:
            # ... let's set a pretty info string with these data
            return f'Raspberry Pi {hardware.group(0)} (Rev. {revision.group(0)})'

        return None

    @staticmethod
    def _fetch_android_device_model() -> Optional[str]:
        """Tries to retrieve `brand` and `model` device properties on Android platforms"""
        try:
            brand = check_output(
                ['getprop', 'ro.product.brand'],
                universal_newlines=True
            ).rstrip()
            model = check_output(
                ['getprop', 'ro.product.model'],
                universal_newlines=True
            ).rstrip()
        except FileNotFoundError:
            return None

        return f'{brand} ({model})'



"""
Simple class (acting as a singleton) dealing with environment variables.
Variables that should alter Archey _global_ behavior may be DRY-ed here.
"""

import os
import platform

from archey.singleton import Singleton


class Environment(metaclass=Singleton):
    """
    At startup, instantiate this class and set up some attributes
      according to their respective environment variable value.
    """
    # See <https://no-color.org/>.
    NO_COLOR = ('NO_COLOR' in os.environ)

    # See <https://bixense.com/clicolors/>.
    CLICOLOR = (os.getenv('CLICOLOR') != '0')
    CLICOLOR_FORCE = (os.getenv('CLICOLOR_FORCE', '0') != '0')

    # See <https://consoledonottrack.com/>.
    DO_NOT_TRACK = (os.getenv('DO_NOT_TRACK') == '1')

    def __init__(self):
        if platform.system() == 'Darwin':
            # Makes future `platform.mac_ver` calls not being _trolled_ by Darwin's kernel
            #   when opening `/System/Library/CoreServices/SystemVersion.plist` file.
            # "Fortunately" for us, `platform` module does not cache these very results.
            # See `platform._mac_ver_xml` function and
            #   <https://eclecticlight.co/2020/08/13/macos-version-numbering-isnt-so-simple/>.
            os.environ['SYSTEM_VERSION_COMPAT'] = '0'



"""Test module for Archey's distribution detection module"""

import unittest
from unittest.mock import MagicMock, patch

from archey.configuration import DEFAULT_CONFIG
from archey.entries.distro import Distro
from archey.test.entries import HelperMethods


class TestDistroEntry(unittest.TestCase):
    """`Distro` entry simple test cases"""
    @patch(
        'archey.entries.distro.check_output',
        return_value='10\n'  # Imitate `getprop` output on Android 10.
    )
    def test_fetch_android_release(self, _):
        """Test `_fetch_android_release` static method"""
        self.assertEqual(
            Distro._fetch_android_release(),  # pylint: disable=protected-access
            'Android 10'
        )

    @patch(
        'archey.entries.distro.platform.mac_ver',
        side_effect=[
            ('', ('', '', ''), ''),                    # Darwin case.
            ('11.1', ('foo', 'bar', 'baz'), 'x86_64')  # macOS case.
        ]
    )
    @patch(
        'archey.entries.distro.platform.release',
        return_value='20.2.0'  # Darwin release.
    )
    def test_fetch_darwin_release(self, _, __):
        """Test `_fetch_darwin_release` static method"""
        self.assertEqual(
            Distro._fetch_darwin_release(),  # pylint: disable=protected-access
            'Darwin 20.2.0'
        )
        self.assertEqual(
            Distro._fetch_darwin_release(),  # pylint: disable=protected-access
            'macOS 11.1'
        )

    @HelperMethods.patch_clean_configuration
    def test_unknown_distro_output(self):
        """Test for `output` method when distribution name couldn't be found"""
        distro_intance_mock = HelperMethods.entry_mock(Distro)
        output_mock = MagicMock()

        distro_intance_mock.value = {
            'name': None,
            'arch': 'ARCHITECTURE'
        }

        Distro.output(distro_intance_mock, output_mock)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            f"{DEFAULT_CONFIG['default_strings']['not_detected']} ARCHITECTURE"
        )


if __name__ == '__main__':
    unittest.main()



"""Test module for Archey's temperature detection module"""

import os
import tempfile

from subprocess import CalledProcessError

import unittest
from unittest.mock import MagicMock, Mock, patch

from archey.configuration import DEFAULT_CONFIG
from archey.entries.temperature import Temperature
from archey.test.entries import HelperMethods
from archey.test import CustomAssertions


class TestTemperatureEntry(unittest.TestCase, CustomAssertions):
    """This module verifies Archey temperature detection"""
    def setUp(self):
        self.temperature_mock = HelperMethods.entry_mock(Temperature)
        self.temperature_mock._temps = []  # pylint: disable=protected-access

    @patch('archey.entries.temperature.run')  # Mock the `sensors` call.
    def test_run_sensors_ok(self, run_mock):
        """Test computations around `sensors` output"""
        run_mock.return_value.stdout = """\
{
   "who-cares-about":{
      "temp1":{
         "temp1_input": 45.000,
         "temp1_crit": 128.000
      },
      "temp2":{
         "temp2_input": 0.000,
         "temp2_crit": 128.000
      },
      "temp3":{
         "temp3_input": 38.000,
         "temp3_crit": 128.000
      },
      "temp4":{
         "temp4_input": 39.000,
         "temp4_crit": 128.000
      },
      "temp5":{
         "temp5_input": 0.000,
         "temp5_crit": 128.000
      },
      "temp6":{
         "temp6_input": 114.000,
         "temp6_crit": 128.000
      }
   },
   "the-chipsets-names":{
      "what-are":{
         "temp1_input": 45.000,
         "temp1_max": 100.000,
         "temp1_crit": 100.000,
         "temp1_crit_alarm": 0.000
      },
      "those":{
         "temp2_input": 43.000,
         "temp2_max": 100.000,
         "temp2_crit": 100.000,
         "temp2_crit_alarm": 0.000
      },
      "identifiers":{
         "temp3_input": 44.000,
         "temp3_max": 100.000,
         "temp3_crit": 100.000,
         "temp3_crit_alarm": 0.000
      }
   },
   "crap-a-fan-chip":{
      "fan1":{
         "fan1_input": 3386.000
      }
   }
}
"""
        # pylint: disable=protected-access
        Temperature._run_sensors(self.temperature_mock, [])
        self.assertListEqual(
            self.temperature_mock._temps,
            [45.0, 38.0, 39.0, 114.0, 45.0, 43.0, 44.0]
        )
        # pylint: enable=protected-access

    @patch('archey.entries.temperature.run')
    def test_run_sensors_ok_multiple_chipsets(self, run_mock):
        """Test `sensors` when multiple chipsets names have been passed"""
        run_mock.side_effect = [
            Mock(
                stdout="""\
{
   "who-cares-about":{
      "temp1":{
         "temp1_input": 45.000,
         "temp1_crit": 128.000
      },
      "temp2":{
         "temp2_input": 0.000,
         "temp2_crit": 128.000
      },
      "temp3":{
         "temp3_input": 38.000,
         "temp3_crit": 128.000
      }
   }
}
""",
                stderr=None
            ),
            Mock(
                stdout="""\
{
   "the-chipsets-names":{
      "what-are":{
         "temp1_input": 45.000,
         "temp1_max": 100.000,
         "temp1_crit": 100.000,
         "temp1_crit_alarm": 0.000
      }
    }
}
""",
                stderr=None
            )
        ]

        sensors_chipsets = ['who-cares-about', 'the-chipsets-names']

        # pylint: disable=protected-access
        Temperature._run_sensors(self.temperature_mock, sensors_chipsets)
        self.assertListEqual(self.temperature_mock._temps, [45.0, 38.0, 45.0])
        # pylint: enable=protected-access

        # Check that our `run` mock has been called up to the number of passed chipsets.
        self.assertEqual(run_mock.call_count, len(sensors_chipsets))

    @patch(
        'archey.entries.temperature.run',
        side_effect=[
            CalledProcessError(1, 'sensors'),  # `sensors` will hard fail.
            FileNotFoundError()                # `sensors` won't be available.
        ]
    )
    def test_run_sensors_ko_exec(self, _):
        """Test `sensors` (hard) failure handling"""
        # pylint: disable=protected-access
        Temperature._run_sensors(self.temperature_mock, [])
        self.assertListEmpty(self.temperature_mock._temps)

        Temperature._run_sensors(self.temperature_mock, [])
        self.assertListEmpty(self.temperature_mock._temps)
        # pylint: enable=protected-access

    @patch('archey.entries.temperature.run')
    def test_run_sensors_ko_output(self, run_mock):
        """Test `sensors` (soft) failure handling"""
        # JSON decoding from `sensors` will fail...
        run_mock.return_value.stdout = """\
{
    "Is this JSON valid ?": [
        "You", "should", "look", "twice.",
    ]
}
"""
        # pylint: disable=protected-access
        Temperature._run_sensors(self.temperature_mock, [])
        self.assertListEmpty(self.temperature_mock._temps)
        # pylint: enable=protected-access

    @patch('archey.entries.temperature.iglob')
    def test_poll_thermal_zones(self, iglob_mock):
        """Tests sensor files handling"""

        ## BEGIN PRELUDE ##
        # We'll store there filenames of some temp files mocking those under
        #  `/sys/class/thermal/thermal_zone*/temp`
        tmp_files = []
        for temperature in (  # Fake temperatures
                b'50000',
                b'0',
                b'40000',
                b'50000'
            ):
            with tempfile.NamedTemporaryFile(delete=False) as temp_file:
                temp_file.write(temperature)
                temp_file.seek(0)
                tmp_files.append(temp_file)
        ## END PRELUDE ##

        iglob_mock.return_value = iter([file.name for file in tmp_files])

        # pylint: disable=protected-access
        Temperature._poll_thermal_zones(self.temperature_mock)
        self.assertListEqual(
            self.temperature_mock._temps,
            [50.0, 40.0, 50.0]
        )
        # pylint: enable=protected-access

        ## BEGIN POSTLUDE ##
        for tmp_file in tmp_files:
            tmp_file.close()
            os.remove(tmp_file.name)
        ## END POSTLUDE ##

    @patch(
        'archey.entries.temperature.check_output',
        side_effect=[
            # First case (`iStats` nor `OSX CPU Temp` won't be available).
            FileNotFoundError(),
            FileNotFoundError(),
            # Second case (`iStats` OK).
            '41.125\n',
            # Third case (`iStats` KO, `OSX CPU Temp` OK).
            FileNotFoundError(),
            '61.8 °C\n',
            # Fourth case (`OSX CPU Temp` KO, but with <= 1.1.0 output).
            # See lavoiesl/osx-cpu-temp#22.
            FileNotFoundError(),
            '0.0°C\n'
        ]
    )
    def test_run_istats_or_osxcputemp(self, _):
        """Test for `iStats` and `OSX CPU Temp` third-party programs outputs"""
        # pylint: disable=protected-access
        # First case.
        Temperature._run_istats_or_osxcputemp(self.temperature_mock)
        self.assertListEmpty(self.temperature_mock._temps)

        # Second case.
        Temperature._run_istats_or_osxcputemp(self.temperature_mock)
        self.assertListEqual(self.temperature_mock._temps, [41.125])

        # Reset internal object here.
        self.temperature_mock._temps = []

        # Third case.
        Temperature._run_istats_or_osxcputemp(self.temperature_mock)
        self.assertListEqual(self.temperature_mock._temps, [61.8])

        # Reset internal object here.
        self.temperature_mock._temps = []

        # Fourth case.
        Temperature._run_istats_or_osxcputemp(self.temperature_mock)
        self.assertListEmpty(self.temperature_mock._temps)
        # pylint: enable=protected-access

    @patch(
        'archey.entries.temperature.os.cpu_count',
        return_value=4  # Mocks a quad-cores CPU system.
    )
    @patch(
        'archey.entries.temperature.check_output',
        side_effect=[
            # First case (`sysctl` won't be available).
            FileNotFoundError(),
            # Second case (`sysctl` will fail).
            CalledProcessError(1, 'sysctl'),
            # Third case (OK).
            """\
42C
42C
45C
45C
""",
            # Fourth case (partially-OK).
            """\
42
40
0
38
"""])
    def test_run_sysctl_dev_cpu(self, _, __):
        """Test for `sysctl` output only"""
        # pylint: disable=protected-access
        # First case.
        Temperature._run_sysctl_dev_cpu(self.temperature_mock)
        self.assertListEmpty(self.temperature_mock._temps)

        # Second case.
        Temperature._run_sysctl_dev_cpu(self.temperature_mock)
        self.assertListEmpty(self.temperature_mock._temps)

        # Third case.
        Temperature._run_sysctl_dev_cpu(self.temperature_mock)
        self.assertListEqual(
            self.temperature_mock._temps,
            [42.0, 42.0, 45.0, 45.0]
        )

        # Reset internal object here.
        self.temperature_mock._temps = []

        # Fourth case.
        Temperature._run_sysctl_dev_cpu(self.temperature_mock)
        self.assertListEqual(
            self.temperature_mock._temps,
            [42.0, 40.0, 38.0]
        )
        # pylint: enable=protected-access

    @patch(
        'archey.entries.temperature.check_output',
        side_effect=[
            FileNotFoundError(),
            'temp=42.8\'C\n'
        ]
    )
    def test_run_vcgencmd(self, _):
        """Test for `vcgencmd` output only"""
        # pylint: disable=protected-access
        Temperature._run_vcgencmd(self.temperature_mock)
        self.assertListEmpty(self.temperature_mock._temps)

        Temperature._run_vcgencmd(self.temperature_mock)
        self.assertListEqual(self.temperature_mock._temps, [42.8])
        # pylint: enable=protected-access

    @HelperMethods.patch_clean_configuration
    def test_output(self):
        """Test `output` method"""
        output_mock = MagicMock()

        # No value --> not detected.
        Temperature.output(self.temperature_mock, output_mock)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            DEFAULT_CONFIG['default_strings']['not_detected']
        )

        output_mock.reset_mock()

        # Values --> normal behavior.
        self.temperature_mock._temps = [  # pylint: disable=protected-access
            50.0, 40.0, 50.0
        ]
        self.temperature_mock.value = {
            'temperature': 46.7,
            'max_temperature': 50.0,
            'char_before_unit': 'o',
            'unit': 'C'
        }
        Temperature.output(self.temperature_mock, output_mock)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            '46.7oC (Max. 50.0oC)'
        )

        # Only one value --> no maximum.
        self.temperature_mock._temps = [42.8]  # pylint: disable=protected-access
        self.temperature_mock.value = {
            'temperature': 42.8,
            'max_temperature': 42.8,
            'char_before_unit': ' ',
            'unit': 'C'
        }
        Temperature.output(self.temperature_mock, output_mock)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            '42.8 C'
        )

    def test_convert_to_fahrenheit(self):
        """Simple tests for the `_convert_to_fahrenheit` static method"""
        test_conversion_cases = (
            (-273.15, -459.67),
            (0.0, 32.0),
            (21.0, 69.8),
            (37.0, 98.6),
            (100.0, 212.0)
        )
        for celsius_value, expected_fahrenheit_value in test_conversion_cases:
            self.assertAlmostEqual(
                Temperature._convert_to_fahrenheit(celsius_value),  # pylint: disable=protected-access
                expected_fahrenheit_value
            )


if __name__ == '__main__':
    unittest.main()



"""Temperature detection class"""

import json
import os
import platform
import re

from glob import iglob
from subprocess import CalledProcessError, DEVNULL, PIPE, check_output, run
from typing import List, Optional

from archey.entry import Entry


class Temperature(Entry):
    """
    Tries to compute an average temperature from `sensors` (LM-Sensors).
    If not available, falls back on system thermal zones files (GNU/Linux)
      or `sysctl` output for BSD and derivatives systems.
    On Raspberry devices, retrieves temperature from the `vcgencmd` binary.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._temps = []

        # Tries `sensors` at first.
        self._run_sensors(self.options.get('sensors_chipsets', []))

        # On error (list still empty)...
        if not self._temps:
            if platform.system() == 'Linux':
                # ... checks for system thermal zones files on GNU/Linux.
                self._poll_thermal_zones()
            elif platform.system() == 'Darwin':
                self._run_istats_or_osxcputemp()
            else:
                # ... or tries `sysctl` calls (available on BSD and derivatives).
                self._run_sysctl_dev_cpu()

        # Tries `vcgencmd` for Raspberry devices.
        self._run_vcgencmd()

        # No value could be fetched, leave `self.value` to `None`.
        if not self._temps:
            return

        # Let's DRY some constants once.
        use_fahrenheit = self.options.get('use_fahrenheit')
        char_before_unit = self.options.get('char_before_unit', ' ')

        # Conversion to Fahrenheit if needed.
        if use_fahrenheit:
            self._temps = list(map(self._convert_to_fahrenheit, self._temps))

        # Final average and maximum computations.
        self.value = {
            'temperature': float(round(sum(self._temps) / len(self._temps), 1)),
            'max_temperature': float(round(max(self._temps), 1)),
            'char_before_unit': char_before_unit,
            'unit': ('F' if use_fahrenheit else 'C')
        }


    def _run_sensors(self, whitelisted_chips: List[str]):

        def _get_sensors_output(whitelisted_chip: Optional[str]) -> Optional[str]:
            sensors_args = ['sensors', '-A', '-j']
            if whitelisted_chip is not None:
                sensors_args.append(whitelisted_chip)

            # Uses the `sensors` program (from lm-sensors) to interrogate thermal chip-sets.
            try:
                sensors_output = run(
                    sensors_args,
                    universal_newlines=True,
                    stdout=PIPE, stderr=PIPE,
                    check=True
                )
            except FileNotFoundError:
                error_message = None
                return None
            except CalledProcessError as called_process_error:
                error_message = called_process_error.stderr
                return None
            else:
                error_message = sensors_output.stderr
            finally:
                # Log any `sensors` error messages at warning level.
                if error_message:
                    for line in error_message.splitlines():
                        self._logger.warning('[lm-sensors]: %s', line)

            return sensors_output.stdout

        sensors_outputs = []
        if whitelisted_chips:
            # Iterate over each whitelisted chipset and store JSON outputs.
            for whitelisted_chip in whitelisted_chips:
                sensors_outputs.append(_get_sensors_output(whitelisted_chip))
        else:
            sensors_outputs.append(_get_sensors_output(None))

        # Loop over output(s) and parse JSON documents.
        for sensors_output in sensors_outputs:
            if sensors_output is None:
                continue

            try:
                sensors_data = json.loads(sensors_output)
            except json.JSONDecodeError as json_decode_error:
                self._logger.warning(
                    'Couldn\'t decode JSON from sensors output : %s', json_decode_error
                )
                continue

            # Iterate over the chipsets outputs to retrieve interesting values.
            for features in sensors_data.values():
                for subfeatures in features.values():
                    for name, value in subfeatures.items():
                        # These conditions check whether this sub-feature value is a correct
                        #  temperature, as :
                        # * It might be an input fan speed (from a control chip) ;
                        # * Some chips/adapters might return null temperatures.
                        if value != 0.0 and re.match(r"temp\d_input", name):
                            self._temps.append(value)
                            # As there is only one `temp*_input` field per item,
                            #  we may stop the current sub-feature iteration now.
                            break

    def _poll_thermal_zones(self):
        # We just check for values within files present in the path below.
        for thermal_file in iglob(r'/sys/class/thermal/thermal_zone*/temp'):
            try:
                with open(thermal_file, encoding='ASCII') as file:
                    temp = float(file.read())
            except OSError:
                continue

            if temp != 0.0:
                self._temps.append(temp / 1000)

    def _run_istats_or_osxcputemp(self):
        """
        For Darwin systems, let's rely on `iStats` or `OSX CPU Temp` third-party programs.
        System's `powermetrics` program is **very** slow to run
          and requires administrator privileges.
        """
        # Run iStats binary (<https://github.com/Chris911/iStats>).
        try:
            istats_output = check_output(
                ['istats', 'cpu', 'temperature', '--value-only'],
                universal_newlines=True
            )
        except OSError:
            pass
        else:
            self._temps.append(float(istats_output))
            return

        # Run OSX CPU Temp binary (<https://github.com/lavoiesl/osx-cpu-temp>).
        try:
            osxcputemp_output = check_output('osx-cpu-temp', universal_newlines=True)
        except OSError:
            pass
        else:
            # Parse output across  <= 1.1.0 and above.
            temp = float(re.search(r'\d+\.\d', osxcputemp_output).group(0))
            if temp != 0.0:  # (Apple) System Management Control read _may_ fail.
                self._temps.append(temp)

    def _run_sysctl_dev_cpu(self):
        # Tries to get temperatures from each CPU core sensor.
        try:
            sysctl_output = check_output(
                ['sysctl', '-n'] + \
                    [f'dev.cpu.{i}.temperature' for i in range(os.cpu_count())],
                stderr=PIPE, universal_newlines=True
            )
        except FileNotFoundError:
            # `sysctl` does not seem to be available on this system.
            return
        except CalledProcessError as error_message:
            self._logger.warning(
                '[sysctl]: Couldn\'t fetch temperature from CPU sensors (%s). '
                'Please be sure to load the corresponding kernel driver beforehand '
                '(`kldload coretemp` for Intel or `kldload amdtemp` for AMD`).',
                (error_message.stderr or 'unknown error').rstrip()
            )
            return

        for temp in sysctl_output.splitlines():
            # Strip any temperature unit from output (some drivers may add it).
            temp = float(temp.rstrip('C'))
            if temp != 0.0:
                self._temps.append(temp)

    def _run_vcgencmd(self):
        # Let's try to retrieve a value from the Broadcom chip on Raspberry.
        try:
            vcgencmd_output = check_output(
                ['/opt/vc/bin/vcgencmd', 'measure_temp'],
                stderr=DEVNULL, universal_newlines=True
            )
        except (FileNotFoundError, CalledProcessError):
            return

        self._temps.append(
            float(
                re.search(
                    r'\d+\.\d+',
                    vcgencmd_output
                ).group(0)
            )
        )

    @staticmethod
    def _convert_to_fahrenheit(temp: float) -> float:
        """Simple Celsius to Fahrenheit conversion method"""
        return temp * (9 / 5) + 32


    def output(self, output):
        """Adds the entry to `output` after pretty-formatting with units."""
        if not self.value:
            # Fall back on the default behavior if no temperatures were detected.
            super().output(output)
            return

        # DRY some constants
        char_before_unit = self.value['char_before_unit']
        unit = self.value['unit']

        entry_text = f"{self.value['temperature']}{char_before_unit}{unit}"
        # When there are multiple input sources, show the hottest value.
        if len(self._temps) > 1:
            entry_text += f" (Max. {self.value['max_temperature']}{char_before_unit}{unit})"

        output.append(self.name, entry_text)



"""Colors enumeration definition"""

from bisect import bisect
from enum import Enum
from functools import lru_cache

import re
import sys

from archey.environment import Environment


# REGEXP compiled pattern matching ANSI/ECMA-48 color escape codes.
ANSI_ECMA_REGEXP = re.compile(r'\x1b\[\d+(?:(?:;\d+)+)?m')


class Colors(Enum):
    """
    ANSI terminal colors enumeration.
    Supports an arbitrary number of display attributes.

    See <http://www.termsys.demon.co.uk/vtansi.htm#colors>.
    """
    CLEAR = (0,)
    RED_NORMAL = (0, 31)
    RED_BRIGHT = (1, 31)
    GREEN_NORMAL = (0, 32)
    GREEN_BRIGHT = (1, 32)
    YELLOW_NORMAL = (0, 33)
    YELLOW_BRIGHT = (1, 33)
    BLUE_NORMAL = (0, 34)
    BLUE_BRIGHT = (1, 34)
    MAGENTA_NORMAL = (0, 35)
    MAGENTA_BRIGHT = (1, 35)
    CYAN_NORMAL = (0, 36)
    CYAN_BRIGHT = (1, 36)
    WHITE_NORMAL = (0, 37)
    WHITE_BRIGHT = (1, 37)

    def __str__(self):
        return self.escape_code_from_attrs(
            ';'.join(map(str, self.value))
        )

    @staticmethod
    @lru_cache(maxsize=None)  # Python < 3.9, `functools.cache` is not yet available.
    def should_color_output() -> bool:
        """
        Returns whether or not output should be colored, according to runtime environment.
        Current implementation is specific to Archey as it's not standardized (see jcs/no_color#28).
        """
        if Environment.CLICOLOR_FORCE:
            return True

        if Environment.NO_COLOR:
            return False

        return (
            sys.stdout.isatty() and Environment.CLICOLOR
        )

    @classmethod
    def escape_code_from_attrs(cls, display_attrs: str) -> str:
        """
        Build and return an ANSI/ECMA-48 escape code string from passed display attributes.
        """
        if not cls.should_color_output():
            return ''

        return f'\x1b[{display_attrs}m'

    @staticmethod
    def get_level_color(value: float, yellow_bpt: float, red_bpt: float) -> 'Colors':
        """Returns the best level color according to `value` compared to `{yellow,red}_bpt`"""
        level_colors = (Colors.GREEN_NORMAL, Colors.YELLOW_NORMAL, Colors.RED_NORMAL)

        return level_colors[bisect((yellow_bpt, red_bpt), value)]

    @staticmethod
    def remove_colors(string: str) -> str:
        """Simple DRY method to remove any ANSI/ECMA-48 color escape code from passed `string`"""
        return ANSI_ECMA_REGEXP.sub('', string)



"""User session detection class"""

import getpass

from archey.entry import Entry


class User(Entry):
    """Retrieves the session name of the current logged in user"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        try:
            self.value = getpass.getuser()
        except ImportError:
            # From <https://github.com/python/cpython/blob/3.9/Lib/getpass.py#L167>,
            #  `pwd` module import _might_ fail.
            pass



"""Elementary OS logo"""

from archey.colors import Colors


COLORS = [Colors.WHITE_BRIGHT]

LOGO = [
    """{c[0]}          eeeeeeeeeeeeeeeee         """,
    """{c[0]}       eeeeeeeeeeeeeeeeeeeeeee      """,
    """{c[0]}     eeeee  eeeeeeeeeeee   eeeee    """,
    """{c[0]}   eeee   eeeee       eee     eeee  """,
    """{c[0]}  eeee   eeee          eee     eeee """,
    """{c[0]} eee    eee            eee       eee""",
    """{c[0]} eee   eee            eee        eee""",
    """{c[0]} ee    eee           eeee       eeee""",
    """{c[0]} ee    eee         eeeee      eeeeee""",
    """{c[0]} ee    eee       eeeee      eeeee ee""",
    """{c[0]} eee   eeee   eeeeee      eeeee  eee""",
    """{c[0]} eee    eeeeeeeeee     eeeeee    eee""",
    """{c[0]}  eeeeeeeeeeeeeeeeeeeeeeee    eeeee """,
    """{c[0]}   eeeeeeee eeeeeeeeeeee      eeee  """,
    """{c[0]}     eeeee                 eeeee    """,
    """{c[0]}       eeeeeee         eeeeeee      """,
    """{c[0]}          eeeeeeeeeeeeeeeee         """
]



"""Alpine Linux logo"""

from archey.colors import Colors


COLORS = [Colors.BLUE_BRIGHT]

LOGO = [
    """{c[0]}        .hddddddddddddddddddddddh.       """,
    """{c[0]}       :dddddddddddddddddddddddddd:      """,
    """{c[0]}      /dddddddddddddddddddddddddddd/     """,
    """{c[0]}     +dddddddddddddddddddddddddddddd+    """,
    """{c[0]}   `sdddddddddddddddddddddddddddddddds`  """,
    """{c[0]}  `ydddddddddddd++hdddddddddddddddddddy` """,
    """{c[0]} .hddddddddddd+`  `+ddddh:-sdddddddddddh.""",
    """{c[0]} hdddddddddd+`      `+y:    .sddddddddddh""",
    """{c[0]} ddddddddh+`   `//`   `.`     -sddddddddd""",
    """{c[0]} ddddddh+`   `/hddh/`   `:s-    -sddddddd""",
    """{c[0]} ddddh+`   `/+/dddddh/`   `+s-    -sddddd""",
    """{c[0]} ddd+`   `/o` :dddddddh/`   `oy-    .yddd""",
    """{c[0]} hdddyo+ohddyosdddddddddho+oydddy++ohdddh""",
    """{c[0]} .hddddddddddddddddddddddddddddddddddddh.""",
    """{c[0]}  `yddddddddddddddddddddddddddddddddddy` """,
    """{c[0]}   `sdddddddddddddddddddddddddddddddds`  """,
    """{c[0]}     +dddddddddddddddddddddddddddddd+    """,
    """{c[0]}      /dddddddddddddddddddddddddddd/     """,
    """{c[0]}       :dddddddddddddddddddddddddd:      """,
    """{c[0]}        .hddddddddddddddddddddddh.       """
]



"""Archey configuration module"""

from copy import deepcopy

import json
import logging
import os

from archey.singleton import Singleton
from archey.utility import Utility


# Below are default required configuration keys which will be used.
DEFAULT_CONFIG = {
    'allow_overriding': True,
    'parallel_loading': True,
    'suppress_warnings': False,
    'honor_ansi_color': True,
    'default_strings': {
        'latest': 'latest',
        'available': 'available',
        'no_address': 'No Address',
        'not_detected': 'Not detected',
        'virtual_environment': 'Virtual Environment'
    }
}


class Configuration(metaclass=Singleton):
    """
    Values present in `DEFAULT_CONFIG` dictionary are required.
    New optional values may be added with `Utility.update_recursive` method.

    If a `config_path` is passed during instantiation, it will be loaded.
    """
    def __init__(self, config_path=None):
        # Deep-copy `DEFAULT_CONFIG` so we have a local copy to safely mutate.
        self._config = deepcopy(DEFAULT_CONFIG)

        # If a `config_path` has been specified, (try to) load it directly.
        if config_path:
            self._load_configuration(config_path)
        # If not, load each (optional) configuration file in a "regular" order.
        else:
            self._load_configuration('/etc/archey4/')
            self._load_configuration(os.path.expanduser('~/.config/archey4/'))
            self._load_configuration(os.getcwd())

    def get(self, key: str, default=None):
        """
        A binding method to imitate the `dict.get()` behavior.
        """
        return self._config.get(key, default)

    def _load_configuration(self, path: str):
        """
        A method handling configuration loading from a JSON file.
        It will try to load any `config.json` present under `path`.
        """
        # If a previous configuration file has denied overriding...
        if not self.get('allow_overriding'):
            #  ... don't load this one.
            return

        # If the specified `path` is a directory, append the file name we are looking for.
        if os.path.isdir(path):
            path = os.path.join(path, 'config.json')

        try:
            with open(path, mode='rb') as f_config:
                Utility.update_recursive(self._config, json.load(f_config))
        except FileNotFoundError:
            return
        except json.JSONDecodeError as json_decode_error:
            logging.error('%s (%s)', json_decode_error, path)
            return

        # When `suppress_warnings` is set, higher the log level to silence warning messages.
        logging.getLogger().setLevel(
            logging.ERROR if self.get('suppress_warnings') else logging.WARN
        )

    def __iter__(self):
        """When used as an iterator, directly yield `_config` elements"""
        return iter(self._config.items())



"""Public IP address detection class"""

from socket import timeout as SocketTimeoutError
from subprocess import check_output, DEVNULL, TimeoutExpired, CalledProcessError
from typing import Optional
from urllib.error import URLError
from urllib.request import urlopen

from archey.entry import Entry
from archey.environment import Environment


class WanIP(Entry):
    """Uses different ways to retrieve the public IPv{4,6} addresses"""
    _PRETTY_NAME = 'WAN IP'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.value = []

        if Environment.DO_NOT_TRACK:
            return

        ipv4_addr = self._retrieve_ip_address(4)
        if ipv4_addr:
            self.value.append(ipv4_addr)

        ipv6_addr = self._retrieve_ip_address(6)
        if ipv6_addr:
            self.value.append(ipv6_addr)


    def _retrieve_ip_address(self, ip_version: int) -> Optional[str]:
        """
        Best effort to retrieve public IP address based on corresponding options.
        We are trying special DNS resolutions first for performance and (system) caching purposes.
        """
        options = self.options.get(f'ipv{ip_version}', {})

        # Is retrieval enabled for this IP version ?
        if not options and not isinstance(options, dict):
            return None

        # Is retrieval via DNS query enabled ?
        dns_query = options.get('dns_query', 'myip.opendns.com')
        if dns_query:
            # Run the DNS query.
            try:
                ip_address = self._run_dns_query(
                    dns_query,
                    options.get('dns_resolver', 'resolver1.opendns.com'),
                    ip_version,
                    options.get('dns_timeout', 1)
                )
            except FileNotFoundError:
                # DNS lookup tool does not seem to be available.
                pass
            else:
                return ip_address

        # Is retrieval via HTTP(S) request enabled ?
        http_url = options.get('http_url', f'https://v{ip_version}.ident.me/')
        if not http_url:
            return None

        # Run the HTTP(S) request.
        return self._run_http_request(
            http_url,
            options.get('http_timeout', 1)
        )


    @staticmethod
    def _run_dns_query(query: str, resolver: str, ip_version: int, timeout: float) -> Optional[str]:
        """Simple wrapper to `dig` command to perform DNS queries"""
        try:
            ip_address = check_output(
                [
                    'dig', '+short',
                    ('-' + str(ip_version)),
                    ('AAAA' if ip_version == 6 else 'A'),
                    query, '@' + resolver
                ],
                timeout=timeout,
                stderr=DEVNULL, universal_newlines=True
            ).rstrip()
        except (TimeoutExpired, CalledProcessError):
            return None

        # `ip_address` might be empty here.
        return ip_address

    @staticmethod
    def _run_http_request(server_url: str, timeout: float) -> Optional[str]:
        """Simple wrapper to `urllib` module to perform HTTP requests"""
        try:
            with urlopen(server_url, timeout=timeout) as http_request:
                return http_request.read().decode().strip()
        except (URLError, SocketTimeoutError):
            return None

    def output(self, output):
        """Adds the entry to `output` after pretty-formatting our list of IP addresses."""
        # If we found IP addresses, join them together nicely.
        # If not, fall-back on the "No address" string.
        if self.value:
            if not self.options.get('one_line', True):
                # One-line output has been disabled, add one IP address per item.
                for ip_address in self.value:
                    output.append(self.name, ip_address)

                return

            text_output = ', '.join(self.value)

        elif not Environment.DO_NOT_TRACK:
            text_output = self._default_strings.get('no_address')
        else:
            text_output = self._default_strings.get('not_detected')

        output.append(self.name, text_output)



"""Distribution and architecture detection class"""

import platform

from subprocess import check_output
from typing import Optional

from archey.distributions import Distributions
from archey.entry import Entry


class Distro(Entry):
    """Uses `distro` and `platform` modules to retrieve distribution and architecture information"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if platform.system() == 'Darwin':
            distro_name = self._fetch_darwin_release()
        else:
            distro_name = Distributions.get_distro_name() or \
                self._fetch_android_release()

        self.value = {
            'name': distro_name,
            'arch': platform.machine()
        }

    @staticmethod
    def _fetch_android_release() -> Optional[str]:
        """Simple method to fetch current release on Android systems"""
        try:
            release = check_output(
                ['getprop', 'ro.build.version.release'],
                universal_newlines=True
            ).rstrip()
        except FileNotFoundError:
            return None

        return f'Android {release}'

    @staticmethod
    def _fetch_darwin_release() -> Optional[str]:
        """Simple method to fetch current release on Darwin systems"""
        # For macOS, let's mimic Python's `platform.platform` internal behavior here.
        macos_release = platform.mac_ver()[0]
        if macos_release:
            return f'macOS {macos_release}'

        return f'Darwin {platform.release()}'

    def output(self, output):
        output.append(
            self.name,
            f"{{}} {self.value['arch']}".format(
                self.value['name'] or self._default_strings.get('not_detected')
            )
        )



"""Simple module doing its best at taking a screenshot of the current screen"""

import logging
import os
import platform
import time

from contextlib import ExitStack
from datetime import datetime
from functools import partial
from subprocess import CalledProcessError, DEVNULL, check_call


def take_screenshot(output_file: str = None) -> bool:
    """
    Simple function trying to take a screenshot using various famous back-end programs.
    When supported by a found and available back-end, **try to** honor `output_file`.
    Returns a `bool` representing whether or not a screenshot could be taken.
    """
    if not output_file or os.path.isdir(output_file):
        # When a directory is provided, we've to force `output_file` to represent a **file** path.
        output_file = os.path.join(
            (output_file or os.getcwd()),
            datetime.now().strftime('archey4_screenshot_%Y-%m-%d_%H.%M.%S.png')
        )

    # Some programs don't accept specific filename as parameters.
    # In such cases, we may provide them a target directory instead.
    output_dir = os.path.dirname(output_file)

    # Back-end programs that _may_ (?) be available across different platforms.
    screenshot_tools = {
        'Flameshot': ['flameshot', 'full', '-p', output_dir],
        'ImageMagick': ['import', '-window', 'root', output_file],
        'maim': ['maim', output_file],
        'scrot': ['scrot', '-z', output_file],
        'Shutter': ['shutter', '-f', '-o', output_file, '-e']
    }

    # Extends the original screenshot tools dictionary according to current platform.
    if platform.system() == 'Windows':
        screenshot_tools['SnippingTool'] = ['SnippingTool.exe', '/clip']
    elif platform.system() == 'Darwin':
        screenshot_tools['ScreenCapture'] = [
            'screencapture',
            '-x',
            '-t', output_file.rpartition('.')[2],
            output_file
        ]
    else:  # *NIX systems (and others)...
        screenshot_tools['Escrotum'] = ['escrotum', output_file]
        screenshot_tools['GNOME-Screenshot'] = ['gnome-screenshot', '-f', output_file]
        screenshot_tools['grim'] = ['grim', output_file]
        screenshot_tools['KDE-Spectacle'] = ['spectacle', '-b', '-o', output_file]
        screenshot_tools['Xfce4-Screenshooter'] = ['xfce4-screenshooter', '-f', '-s', output_dir]
        screenshot_tools['Screencap (Android)'] = [
            'screencap',  # Binary available on Android.
            '-p',         # It only accepts PNG as image output format.
            (output_file.rpartition('.')[0] + '.png')
        ]

    # This part purposefully blocks so we wait a little bit before taking the screenshot.
    # It prevents taking a screenshot before Archey's output has appeared.
    for time_remaining in range(3, 0, -1):
        taking_sc_str = f'Taking screenshot in {time_remaining:1d}...'
        print(taking_sc_str, end='', flush=True)
        time.sleep(1)
        print('\r' + ' ' * len(taking_sc_str), end='\r', flush=True)
    time.sleep(0.5)

    with ExitStack() as defer_stack:
        for screenshot_tool, screenshot_cmd in screenshot_tools.items():
            try:
                check_call(screenshot_cmd, stderr=DEVNULL)
            except OSError:
                continue
            except CalledProcessError as process_error:
                defer_stack.callback(partial(
                    logging.warning,
                    'Couldn\'t take a screenshot with %s: \"%s\".', screenshot_tool, process_error
                ))
                continue

            return True

    logging.error(
        """\
Sorry, we couldn\'t find any supported program to take a screenshot on your system.
Please install one of the following and try again: %s.""",
        ', '.join(screenshot_tools.keys())
    )
    return False



"""Test module for Archey's device's model detection module"""

from subprocess import CalledProcessError

import unittest
from unittest.mock import MagicMock, mock_open, patch

from archey.configuration import DEFAULT_CONFIG
from archey.entries.model import Model
from archey.test.entries import HelperMethods


class TestModelEntry(unittest.TestCase):
    """
    For this test we have to go through several eventualities :
    * Laptop / Desktop "regular" environments
    * Raspberry Pi
    * Virtual environment (as a VM or a container)
    * Android devices
    """
    @patch('archey.entries.model.check_output')
    @HelperMethods.patch_clean_configuration
    def test_fetch_virtual_env_info(self, check_output_mock):
        """Test `_fetch_virtual_env_info` method"""
        model_mock = HelperMethods.entry_mock(Model)

        with self.subTest('Detected virtual environment.'):
            check_output_mock.side_effect = [
                FileNotFoundError(),  # `systemd-detect-virt` is not available.
                'xen\nxen-domU\n',    # `virt-what` example output.
                'HYPERVISOR-NAME\n'   # `dmidecode` example output.
            ]

            self.assertEqual(
                Model._fetch_virtual_env_info(model_mock),  # pylint: disable=protected-access
                'HYPERVISOR-NAME (xen, xen-domU)'
            )

        check_output_mock.reset_mock()

        with self.subTest('Virtual environment without `dmidecode`.'):
            check_output_mock.side_effect = [
                FileNotFoundError(),  # `systemd-detect-virt` is not available.
                'xen\nxen-domU\n',    # `virt-what` example output.
                FileNotFoundError()   # `dmidecode` will fail.
            ]

            self.assertEqual(
                Model._fetch_virtual_env_info(model_mock),  # pylint: disable=protected-access
                f"{DEFAULT_CONFIG['default_strings']['virtual_environment']} (xen, xen-domU)"
            )

        check_output_mock.reset_mock()

        with self.subTest('Virtual environment with systemd only.'):
            check_output_mock.side_effect = [
                'systemd-nspawn\n',  # `systemd-detect-virt` output.
                PermissionError()    # `dmidecode` will fail.
            ]

            self.assertEqual(
                Model._fetch_virtual_env_info(model_mock),  # pylint: disable=protected-access
                f"{DEFAULT_CONFIG['default_strings']['virtual_environment']} (systemd-nspawn)"
            )

        check_output_mock.reset_mock()

        with self.subTest('Virtual environment with systemd and `dmidecode`.'):
            check_output_mock.side_effect = [
                'systemd-nspawn\n',  # `systemd-detect-virt` example output.
                                     # `virt-what` won't be called (systemd call succeeded).
                'HYPERVISOR-NAME\n'  # `dmidecode` example output.
            ]

            self.assertEqual(
                Model._fetch_virtual_env_info(model_mock),  # pylint: disable=protected-access
                'HYPERVISOR-NAME (systemd-nspawn)'
            )

        check_output_mock.reset_mock()

        with self.subTest('Not a virtual environment (systemd).'):
            check_output_mock.side_effect = CalledProcessError(  # pylint: disable=redefined-variable-type
                1, 'systemd-detect-virt', 'none\n'
            )

            self.assertIsNone(
                Model._fetch_virtual_env_info(model_mock)  # pylint: disable=protected-access
            )

        check_output_mock.reset_mock()

        with self.subTest('Not a virtual environment (virt-what).'):
            check_output_mock.side_effect = [
                FileNotFoundError(),  # `systemd-detect-virt` won't be available.
                '\n'                  # `virt-what` won't detect anything.
                                      # `dmidecode` won't even be called.
            ]

            self.assertIsNone(
                Model._fetch_virtual_env_info(model_mock)  # pylint: disable=protected-access
            )

        check_output_mock.reset_mock()

        with self.subTest('Not a virtual environment (no tools or not enough privileges)'):
            check_output_mock.side_effect = [
                FileNotFoundError(),  # `systemd-detect-virt` won't be available.
                PermissionError()     # `virt-what` will fail.
            ]

            self.assertIsNone(
                Model._fetch_virtual_env_info(model_mock)  # pylint: disable=protected-access
            )

    def test_fetch_dmi_info(self):
        """Test `_fetch_dmi_info` static method"""
        # `/sys` could not be read from.
        with patch('archey.entries.model.open', mock_open()) as mock:
            mock.return_value.read.side_effect = [PermissionError(), PermissionError()]

            self.assertIsNone(Model._fetch_dmi_info())  # pylint: disable=protected-access

        # All product information are available.
        with patch('archey.entries.model.open', mock_open()) as mock:
            mock.return_value.read.side_effect = [
                'PRODUCT-NAME\n',
                'PRODCT-VENDOR\n',
                'PRODUCT-VERSION\n'
            ]

            self.assertEqual(
                Model._fetch_dmi_info(),  # pylint: disable=protected-access
                'PRODCT-VENDOR PRODUCT-NAME PRODUCT-VERSION'
            )

        # Only product name and version are available.
        with patch('archey.entries.model.open', mock_open()) as mock:
            mock.return_value.read.side_effect = [
                'PRODUCT-NAME\n',
                FileNotFoundError(),
                'PRODUCT-VERSION\n'
            ]

            self.assertEqual(
                Model._fetch_dmi_info(),  # pylint: disable=protected-access
                'PRODUCT-NAME PRODUCT-VERSION'
            )

        # Product name is not available but some board information are.
        with patch('archey.entries.model.open', mock_open()) as mock:
            mock.return_value.read.side_effect = [
                FileNotFoundError(),
                'BOARD-NAME\n',
                'To Be Filled By O.E.M.\n',
                'BOARD-VERSION\n'
            ]

            self.assertEqual(
                Model._fetch_dmi_info(),  # pylint: disable=protected-access
                'BOARD-NAME BOARD-VERSION'
            )

        # Product name nor board name are available.
        with patch('archey.entries.model.open', mock_open()) as mock:
            mock.return_value.read.side_effect = [
                FileNotFoundError(),
                FileNotFoundError()
            ]

            self.assertIsNone(Model._fetch_dmi_info())  # pylint: disable=protected-access

    @patch(
        'archey.entries.model.platform.system',
        side_effect=[
            'Darwin',
            'Darwin',
            'OpenBSD',
            'OpenBSD',
            'OpenBSD'
        ]
    )
    @patch(
        'archey.entries.model.check_output',
        side_effect=[
            # First case [Darwin] (`sysctl` won't be available).
            FileNotFoundError(),
            # Second case [Darwin] OK.
            'MacBookPro14,2',
            # Fifth case [BSD] (`sysctl` won't be available).
            FileNotFoundError(),
            # Third case [BSD] (OK).
            'VMware, Inc.\n',
            'VMware Virtual Platform\n',
            'None\n',
            # Fourth case [BSD] (partially-OK).
            CalledProcessError(1, 'sysctl'),
            'KALAP10D300EA\n',
            '1\n'
        ]
    )
    def test_fetch_sysctl_hw(self, _, __):
        """Test `_fetch_sysctl_hw` static method"""
        # pylint: disable=protected-access
        # Dawin cases.
        self.assertIsNone(Model._fetch_sysctl_hw())
        self.assertEqual(Model._fetch_sysctl_hw(), 'MacBookPro14.2')
        # BSD cases.
        self.assertIsNone(Model._fetch_sysctl_hw())
        self.assertEqual(
            Model._fetch_sysctl_hw(),
            'VMware, Inc. VMware Virtual Platform'
        )
        self.assertEqual(Model._fetch_sysctl_hw(), 'KALAP10D300EA 1')
        # pylint: enable=protected-access

    def test_fetch_raspberry_pi_revision(self):
        """Test `_fetch_raspberry_pi_revision` static method"""
        with patch('archey.entries.model.open', mock_open()) as mock:
            mock.return_value.read.side_effect = [
                'Hardware\t: HARDWARE\nRevision\t: REVISION\n',
                'processor   : 0\ncpu family  : X\n'
            ]

            self.assertEqual(
                Model._fetch_raspberry_pi_revision(),  # pylint: disable=protected-access
                'Raspberry Pi HARDWARE (Rev. REVISION)'
            )
            self.assertIsNone(
                Model._fetch_raspberry_pi_revision()  # pylint: disable=protected-access
            )

    @patch(
        'archey.entries.model.check_output',
        side_effect=[
            'PHONE-BRAND\n',     # First `getprop` call.
            'PHONE-DEVICE\n',    # Second `getprop` call.
            FileNotFoundError()  # Second test will fail.
        ]
    )
    def test_fetch_android_device_model(self, _):
        """Test `_fetch_android_device_model` static method"""
        self.assertEqual(
            Model._fetch_android_device_model(),  # pylint: disable=protected-access
            'PHONE-BRAND (PHONE-DEVICE)'
        )
        self.assertIsNone(
            Model._fetch_android_device_model()  # pylint: disable=protected-access
        )

    @HelperMethods.patch_clean_configuration
    def test_no_match(self):
        """Test when no information could be retrieved"""
        model_instance_mock = HelperMethods.entry_mock(Model)

        output_mock = MagicMock()
        Model.output(model_instance_mock, output_mock)

        self.assertIsNone(model_instance_mock.value)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            DEFAULT_CONFIG['default_strings']['not_detected']
        )


if __name__ == '__main__':
    unittest.main()



"""Archey API module"""

from datetime import datetime

import json

from typing import List

from archey._version import __version__
from archey.distributions import Distributions
from archey.entry import Entry
from archey.utility import Utility


class API:
    """
    This class provides results serialization for external usages.
    At the moment, only JSON has been implemented.
    Feel free to contribute to add other formats as needed.
    """
    def __init__(self, entries: List[Entry]):
        self.entries = entries

    def json_serialization(self, indent: int = 0) -> str:
        """
        JSON serialization of entries.
        Set `indent` to the number of wanted output indentation tabs (2-space long).
        """
        document = {
            'data': {
                entry.name: entry.value for entry in self.entries
            },
            'meta': {
                'version': Utility.version_to_semver_segments(__version__),
                'date': datetime.now().isoformat(),
                'count': len(self.entries),
                'distro': Distributions.get_local().value,
            }
        }

        return json.dumps(
            document,
            indent=((indent * 2) or None)
        )



"""Test module for Archey's GPU detection module"""

import unittest
from unittest.mock import MagicMock, call, patch

from archey.configuration import DEFAULT_CONFIG
from archey.entries.gpu import GPU
from archey.test import CustomAssertions
from archey.test.entries import HelperMethods


class TestGPUEntry(unittest.TestCase, CustomAssertions):
    """Here, we mock the `check_output` call to `lspci` to test the logic"""
    @patch(
        'archey.entries.gpu.check_output',
        side_effect=[
            FileNotFoundError(),
            """\
XX:YY.H IDE interface: IIIIIIIIIIIIIIII
XX:YY.H SMBus: BBBBBBBBBBBBBBBB
XX:YY.H VGA compatible controller: GPU-MODEL-NAME
XX:YY.H Audio device: DDDDDDDDDDDDDDDD
""",
            """\
XX:YY.H IDE interface: IIIIIIIIIIIIIIII
XX:YY.H SMBus: BBBBBBBBBBBBBBBB
XX:YY.H VGA compatible controller: GPU-MODEL-NAME
XX:YY.H Display controller: ANOTHER-MATCHING-VIDEO-CONTROLLER
XX:YY.H Audio device: DDDDDDDDDDDDDDDD
XX:YY.H 3D controller: 3D GPU-MODEL-NAME TAKES ADVANTAGE
"""])
    def test_parse_lspci_output(self, _):
        """Check `_parse_lspci_output` behavior"""
        # pylint: disable=protected-access
        self.assertListEmpty(GPU._parse_lspci_output())
        self.assertListEqual(GPU._parse_lspci_output(), ['GPU-MODEL-NAME'])
        self.assertListEqual(
            GPU._parse_lspci_output(),
            [
                '3D GPU-MODEL-NAME TAKES ADVANTAGE',
                'GPU-MODEL-NAME',
                'ANOTHER-MATCHING-VIDEO-CONTROLLER'
            ]
        )
        # pylint: enable=protected-access

    @patch(
        'archey.entries.gpu.check_output',
        side_effect=[
            FileNotFoundError(),
            """\
Intel HD Graphics 530:

  Chipset Model: Intel HD Graphics 530
  Type: GPU
  Bus: Built-In
  VRAM (Dynamic, Max): 1536 MB
  Vendor: Intel
  Device ID: 0x191b
  Revision ID: 0x0006
  Automatic Graphics Switching: Supported
  gMux Version: 4.0.29 [3.2.8]
  Metal: Supported, feature set macOS GPUFamily2 v1
  Displays:
    Color LCD:
      Display Type: Built-In Retina LCD
      Resolution: 2880 x 1800 Retina
      Framebuffer Depth: 24-Bit Color (ARGB8888)
      Main Display: Yes
      Mirror: Off
      Online: Yes
      Rotation: Supported
      Automatically Adjust Brightness: No

Radeon Pro 450:

  Chipset Model: AMD Radeon Pro 450
  Type: GPU
  Bus: PCIe
  PCIe Lane Width: x8
  VRAM (Total): 2 GB
  Vendor: AMD (0x1002)
  Device ID: 0x67ef
  Revision ID: 0x00ef
  ROM Revision: 113-C980AF-908
  VBIOS Version: 113-C9801AL-028
  EFI Driver Version: 01.A0.908
  Automatic Graphics Switching: Supported
  gMux Version: 4.0.29 [3.2.8]
  Metal: Supported, feature set macOS GPUFamily2 v1
"""])
    def test_parse_system_profiler(self, _):
        """Check `_parse_system_profiler` behavior"""
        # pylint: disable=protected-access
        self.assertListEmpty(GPU._parse_system_profiler())
        self.assertListEqual(
            GPU._parse_system_profiler(),
            ['Intel HD Graphics 530', 'AMD Radeon Pro 450']
        )
        # pylint: enable=protected-access

    @patch(
        'archey.entries.gpu.check_output',
        side_effect=[
            FileNotFoundError(),
            """\
vgapci0@pci0:0:1:0:   class=0x030000 card=0x21f617aa chip=0x01668086 rev=0x09 hdr=0x00
    vendor     = 'Intel Corporation'
    device     = '3rd Gen Core processor Graphics Controller'
    class      = display
    subclass   = VGA
hostb0@pci0:0:1:2:    class=0x060000 card=0x00000000 chip=0x12378086 rev=0x02 hdr=0x00
    vendor     = 'Intel Corporation'
    device     = '82440/1FX 440FX (Natoma) System Controller'
    class      = bridge
    subclass   = HOST-PCI
isab0@pci0:0:1:4:     class=0x060100 card=0x00000000 chip=0x70008086 rev=0x00 hdr=0x00
    vendor     = 'Intel Corporation'
    device     = 'PIIX3 PCI-to-ISA Bridge (Triton II) (82371SB)'
    class      = bridge
    subclass   = PCI-ISA
atapci0@pci0:0:1:6:   class=0x01018a card=0x00000000 chip=0x71118086 rev=0x01 hdr=0x00
    vendor     = 'Intel Corporation'
    device     = 'PIIX4/4E/4M IDE Controller (82371AB/EB/MB)'
    class      = mass storage
    subclass   = ATA
vgapci0@pci0:1:0:0:   class=0x030000 card=0x84391043 chip=0x0fc610de rev=0xa1 hdr=0x00
    vendor     = 'NVIDIA Corporation'
    device     = 'GK107 [GeForce GTX 650]'
    class      = display
    subclass   = VGA
bge0@pci0:1:0:2:      class=0x020000 card=0x2133103c chip=0x165f14e4 rev=0x00 hdr=0x00
    vendor     = 'Broadcom Corporation'
    device     = 'NetXtreme BCM5720 Gigabit Ethernet PCIe'
    class      = network
    subclass   = ethernet
none1@pci0:14:0:0:    class=0x0c0330 card=0x1996103c chip=0x00141912 rev=0x03 hdr=0x00
    vendor     = 'Renesas Technology Corp.'
    class      = serial bus
    subclass   = USB
vgapci0@pci0:16:0:0:  class=0x030000 card=0x3381103c chip=0x0533102b rev=0x00 hdr=0x00
    vendor     = 'Matrox Graphics, Inc.'
    device     = 'MGA G200EH'
    class      = display
    subclass   = VGA
"""])
    def test_parse_pciconf_output(self, _):
        """Check `_parse_pciconf_output` behavior"""
        # pylint: disable=protected-access
        self.assertListEmpty(GPU._parse_pciconf_output())
        self.assertListEqual(
            GPU._parse_pciconf_output(),
            [
                'Intel Corporation 3rd Gen Core processor Graphics Controller',
                'NVIDIA Corporation GK107 [GeForce GTX 650]',
                'Matrox Graphics, Inc. MGA G200EH'
            ]
        )
        # pylint: enable=protected-access

    @HelperMethods.patch_clean_configuration
    def test_various_output_configuration(self):
        """Test `output` overloading based on user preferences"""
        gpu_instance_mock = HelperMethods.entry_mock(GPU)
        output_mock = MagicMock()

        gpu_instance_mock.value = [
            '3D GPU-MODEL-NAME TAKES ADVANTAGE',
            'GPU-MODEL-NAME',
            'ANOTHER-MATCHING-VIDEO'
        ]

        with self.subTest('Single-line combined output.'):
            gpu_instance_mock.options['one_line'] = True

            GPU.output(gpu_instance_mock, output_mock)
            output_mock.append.assert_called_once_with(
                'GPU',
                '3D GPU-MODEL-NAME TAKES ADVANTAGE, GPU-MODEL-NAME, ANOTHER-MATCHING-VIDEO'
            )

        output_mock.reset_mock()

        with self.subTest('Multi-lines output.'):
            gpu_instance_mock.options['one_line'] = False

            GPU.output(gpu_instance_mock, output_mock)
            self.assertEqual(output_mock.append.call_count, 3)
            output_mock.append.assert_has_calls(
                [
                    call(
                        'GPU',
                        '3D GPU-MODEL-NAME TAKES ADVANTAGE'
                    ),
                    call(
                        'GPU',
                        'GPU-MODEL-NAME'
                    ),
                    call(
                        'GPU',
                        'ANOTHER-MATCHING-VIDEO'
                    )
                ]
            )

        output_mock.reset_mock()

        with self.subTest('No GPU detected output.'):
            gpu_instance_mock.value = []

            GPU.output(gpu_instance_mock, output_mock)
            output_mock.append.assert_called_once_with(
                'GPU',
                DEFAULT_CONFIG['default_strings']['not_detected']
            )


if __name__ == '__main__':
    unittest.main()



"""Test module for Archey's kernel information detection module"""

import unittest
from unittest.mock import MagicMock, Mock, patch

from archey.configuration import DEFAULT_CONFIG
from archey.entries.kernel import Kernel
from archey.test.entries import HelperMethods


class TestKernelEntry(unittest.TestCase):
    """
    Here, we mock the `platform` module calls and check afterwards
      that the output is correct.
    """
    @patch(
        'archey.entries.kernel.platform.system',
        return_value='Linux'
    )
    @patch(
        'archey.entries.kernel.platform.release',
        return_value='X.Y.Z-R-arch'
    )
    def test_fetch_kernel_release(self, _, __):
        """Verify `platform` module mocking"""
        self.assertEqual(
            Kernel().value['name'],
            'Linux'
        )
        self.assertEqual(
            Kernel().value['release'],
            'X.Y.Z-R-arch'
        )

    @patch('archey.entries.kernel.urlopen')
    def test_fetch_latest_linux_release(self, urlopen_mock):
        """Check proper JSON decoding and value gathering"""
        urlopen_mock.return_value.__enter__.return_value.read.return_value = b"""\
{
    "latest_stable": {
        "version": "5.10.1"
    },
    "releases": [
        {
            "iseol": false,
            "version": "5.10",
            "moniker": "mainline",
            "source": "https://cdn.kernel.org/pub/linux/kernel/v5.x/linux-5.10.tar.xz",
            "pgp": "https://cdn.kernel.org/pub/linux/kernel/v5.x/linux-5.10.tar.sign",
            "released": {
                "timestamp": 1607899290,
                "isodate": "2020-12-13"
            },
            "gitweb": "https://git.kernel.org/torvalds/h/v5.10",
            "changelog": null,
            "diffview": "https://git.kernel.org/torvalds/ds/v5.10/v5.9",
            "patch": {
                "full": "https://cdn.kernel.org/pub/linux/kernel/v5.x/patch-5.10.xz",
                "incremental": null
            }
        },
        {
            "iseol": false,
            "version": "5.10.1",
            "moniker": "stable",
            "source": "https://cdn.kernel.org/pub/linux/kernel/v5.x/linux-5.10.1.tar.xz",
            "pgp": "https://cdn.kernel.org/pub/linux/kernel/v5.x/linux-5.10.1.tar.sign",
            "released": {
                "timestamp": 1607970794,
                "isodate": "2020-12-14"
            },
            "gitweb": "https://git.kernel.org/stable/h/v5.10.1",
            "changelog": "https://cdn.kernel.org/pub/linux/kernel/v5.x/ChangeLog-5.10.1",
            "diffview": "https://git.kernel.org/stable/ds/v5.10.1/v5.10",
            "patch": {
                "full": "https://cdn.kernel.org/pub/linux/kernel/v5.x/patch-5.10.1.xz",
                "incremental": null
            }
        }
    ]
}"""
        self.assertEqual(
            Kernel._fetch_latest_linux_release(),  # pylint: disable=protected-access
            '5.10.1'
        )

    @patch(
        'archey.entries.kernel.platform.system',
        return_value='Java'
    )
    @patch(
        'archey.entries.kernel.platform.release',
        return_value='X.Y.Z-R-arch'
    )
    @patch(
        'archey.entries.kernel.Environment',
        Mock(DO_NOT_TRACK=False)
    )
    def test_non_linux_platform(self, _, __):
        """Check behavior on non-Linux platforms"""
        kernel = Kernel(options={'check_version': True})

        self.assertIsNone(kernel.value['latest'])
        self.assertIsNone(kernel.value['is_outdated'])

    @patch(
        'archey.entries.kernel.platform.release',
        return_value='X.Y.Z-R-arch'
    )
    @patch(
        'archey.entries.kernel.Environment',
        Mock(DO_NOT_TRACK=True)
    )
    def test_do_not_track(self, _):
        """Check `DO_NOT_TRACK` is correctly honored"""
        kernel = Kernel(options={'check_version': True})

        self.assertIsNone(kernel.value['latest'])
        self.assertIsNone(kernel.value['is_outdated'])

    @patch(
        'archey.entries.kernel.platform.release',
        return_value='1.2.3-4-arch'
    )
    @patch(
        'archey.entries.kernel.Kernel._fetch_latest_linux_release',
        side_effect=[
            '1.2.3',
            '1.3.2'
        ]
    )
    @patch(
        'archey.entries.kernel.platform.system',
        return_value='Linux'
    )
    @patch(
        'archey.entries.kernel.Environment',
        Mock(DO_NOT_TRACK=False)
    )
    @HelperMethods.patch_clean_configuration
    def test_kernel_comparison(self, _, __, ___):
        """Check kernel releases comparison and output templates"""
        output_mock = MagicMock()

        # Only current release (`check_version` disabled by default).
        kernel = Kernel()
        kernel.output(output_mock)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            'Linux 1.2.3-4-arch'
        )

        # Current = latest (up to date !).
        kernel = Kernel(options={'check_version': True})
        kernel.output(output_mock)

        self.assertTrue(kernel.value['latest'])
        self.assertIs(kernel.value['is_outdated'], False)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            f"Linux 1.2.3-4-arch ({DEFAULT_CONFIG['default_strings']['latest']})"
        )

        # Current < latest (outdated).
        kernel = Kernel(options={'check_version': True})
        kernel.output(output_mock)

        self.assertTrue(kernel.value['latest'])
        self.assertIs(kernel.value['is_outdated'], True)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            f"Linux 1.2.3-4-arch (1.3.2 {DEFAULT_CONFIG['default_strings']['available']})"
        )


if __name__ == '__main__':
    unittest.main()



"""Android logo"""

from archey.colors import Colors


COLORS = [Colors.GREEN_BRIGHT, Colors.WHITE_BRIGHT]

LOGO = [
    """{c[0]}          -o          o-         """,
    """{c[0]}           +hydNNNNdyh+          """,
    """{c[0]}         +mMMMMMMMMMMMMm+        """,
    """{c[0]}       `dMM{c[1]}m:{c[0]}NMMMMMMN{c[1]}:m{c[0]}MMd`      """,
    """{c[0]}       hMMMMMMMMMMMMMMMMMMh      """,
    """{c[0]}   ..  yyyyyyyyyyyyyyyyyyyy  ..  """,
    """{c[0]} .mMMm`MMMMMMMMMMMMMMMMMMMM`mMMm.""",
    """{c[0]} :MMMM-MMMMMMMMMMMMMMMMMMMM-MMMM:""",
    """{c[0]} :MMMM-MMMMMMMMMMMMMMMMMMMM-MMMM:""",
    """{c[0]} :MMMM-MMMMMMMMMMMMMMMMMMMM-MMMM:""",
    """{c[0]} :MMMM-MMMMMMMMMMMMMMMMMMMM-MMMM:""",
    """{c[0]} -MMMM-MMMMMMMMMMMMMMMMMMMM-MMMM-""",
    """{c[0]}  +yy+ MMMMMMMMMMMMMMMMMMMM +yy+ """,
    """{c[0]}       mMMMMMMMMMMMMMMMMMMm      """,
    """{c[0]}       `/++MMMMh++hMMMM++/`      """,
    """{c[0]}           MMMMo  oMMMM          """,
    """{c[0]}           MMMMo  oMMMM          """,
    """{c[0]}           oNMm-  -mMNs          """
]



"""`__init__` file for the `logos` submodule, containing dedicated utility methods"""

from importlib import import_module
from types import ModuleType
from typing import List


def lazy_load_logo_module(logo_name: str) -> ModuleType:
    """
    Utility function returning a logo (as a Python module) lazily-loaded.
    It allows us to only load to RAM the distribution logo object that will actually be used.
    """
    return import_module(f"{__name__}.{logo_name}")


def get_logo_width(logo: List[str], nb_colors: int = 8) -> int:
    """
    Utility function computing the real width of a distribution logo.
    Rationale : We use placeholders to dynamically set ANSI colors.
                Although, they **DO NOT** take width space once the text as been printed.

    For performance purposes we compute the logo length based on its first line.
    See `archey.test.test_archey_logos` unit tests for further logos consistency verifications.

    `logo` is supposed to be one of the constants declared above.
    `nb_colors` must be greater than or equal to the number of colors used by the logo.
    """
    # We replace each placeholder by a 0-character string.
    return len(logo[0].format(c=[''] * nb_colors))



"""Simple class (acting as a singleton) to handle processes listing"""

import logging

from subprocess import CalledProcessError, PIPE, check_output

from archey.singleton import Singleton


class Processes(metaclass=Singleton):
    """At startup, instantiate this class to populate a list of running processes"""
    def __init__(self):
        try:
            ps_output = check_output(
                ['ps', '-eo', 'comm'],
                stderr=PIPE, universal_newlines=True
            )
        except FileNotFoundError:
            self._processes = []
            logging.warning("`procps` (or `procps-ng`) couldn't be found on your system.")
        except CalledProcessError as process_error:
            self._processes = []
            logging.warning(
                "This implementation of `ps` might not be supported : %s", process_error.stderr
            )
        else:
            # Discard first heading line here.
            self._processes = ps_output.splitlines()[1:]

    @property
    def list(self) -> tuple:
        """Simple getter to retrieve (am immutable copy of) the processes list"""
        return tuple(self._processes)

    @property
    def number(self) -> int:
        """Simple getter to retrieve the number of stored processes"""
        return len(self._processes)



"""RAM usage detection class"""

import platform
import re

from subprocess import check_output
from typing import Tuple

from archey.colors import Colors
from archey.exceptions import ArcheyException
from archey.entry import Entry


class RAM(Entry):
    """
    First tries to use the `free` command to retrieve RAM usage.
    If not available, falls back on the parsing of `/proc/meminfo` file.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        used, total = self._get_used_total_values()
        if not total:
            return

        self.value = {
            'used': used,
            'total': total,
            'unit': 'MiB'
        }

    def _get_used_total_values(self) -> Tuple[float, float]:
        """
        Returns a tuple containing used and total RAM values.
        Tries a variety of methods, increasing compatibility for a wide range of systems.
        """
        if platform.system() == 'Linux':
            try:
                return self._run_free_dash_m()
            except (IndexError, FileNotFoundError):
                pass
        else:
            # Darwin or any other BSD-based system.
            try:
                return self._run_sysctl_and_vmstat()
            except FileNotFoundError:
                pass

        try:
            return self._read_proc_meminfo()
        except OSError:
            pass

        return 0, 0

    @staticmethod
    def _run_free_dash_m() -> Tuple[float, float]:
        """Call `free -m` and parse its output to retrieve current used and total RAM"""
        memory_usage = ''.join(
            filter(
                re.compile(r'Mem').search,
                check_output(
                    ['free', '-m'],
                    env={'LANG': 'C'}, universal_newlines=True
                ).splitlines()
            )
        ).split()

        return float(memory_usage[2]), float(memory_usage[1])

    @staticmethod
    def _read_proc_meminfo() -> Tuple[float, float]:
        """Same behavior but by reading from `/proc/meminfo` directly"""
        with open('/proc/meminfo', encoding='ASCII') as f_mem_info:
            mem_info_lines = f_mem_info.read().splitlines()

        # Store memory information into a dictionary.
        mem_info = {}
        for line in filter(None, mem_info_lines):
            key, value = line.split(':', maxsplit=1)
            mem_info[key] = float(value.strip(' kB')) / 1024

        total = mem_info['MemTotal']
        # Here, let's imitate Neofetch's behavior.
        # See <https://github.com/dylanaraps/neofetch/wiki/Frequently-Asked-Questions>.
        used = total + mem_info['Shmem'] - (
            mem_info['MemFree'] + mem_info['Cached']
            + mem_info['SReclaimable'] + mem_info['Buffers']
        )
        # Imitates what `free` does when the obtained value happens to be incorrect.
        # See <https://gitlab.com/procps-ng/procps/blob/master/proc/sysinfo.c#L790>.
        if used < 0:
            used = total - mem_info['MemFree']

        return used, total

    @staticmethod
    def _run_sysctl_and_vmstat() -> Tuple[float, float]:
        """From `sysctl` and `vm_stat` calls, compute used and total system RAM values"""
        total = float(
            check_output(
                ['sysctl', '-n', 'hw.memsize'],
                universal_newlines=True
            )
        )

        vm_stat_lines = check_output('vm_stat', universal_newlines=True).splitlines()

        # From first heading line, fetch system page size.
        page_size_match = re.match(
            r'^Mach Virtual Memory Statistics: \(page size of (\d+) bytes\)$',
            vm_stat_lines[0]
        )
        if not page_size_match:
            raise ArcheyException("Couldn't parse `vm_stat` output, please open an issue.")
        page_size = int(page_size_match.group(1))

        # Store memory information into a dictionary.
        mem_info = {}
        for line in vm_stat_lines[1:]:  # We ignore the first heading line.
            key, value = line.split(':', maxsplit=1)
            mem_info[key] = int(value.lstrip().rstrip('.'))

        # Here we imitate Ansible behavior to compute used RAM.
        used = total - (
            mem_info['Pages wired down'] + mem_info['Pages active'] + mem_info['Pages inactive']
        ) * page_size

        return (used / 1024**2), (total / 1024**2)


    def output(self, output):
        """
        Adds the entry to `output` after pretty-formatting the RAM usage with color and units.
        """
        if not self.value:
            # Fall back on the default behavior if no RAM usage could be detected.
            super().output(output)
            return

        # DRY some constants
        used = self.value['used']
        total = self.value['total']
        unit = self.value['unit']

        # Based on the RAM percentage usage, select the corresponding level color.
        level_color = Colors.get_level_color(
            (used / total) * 100,
            self.options.get('warning_use_percent', 33.3),
            self.options.get('danger_use_percent', 66.7)
        )

        output.append(
            self.name,
            f'{level_color}{int(used)} {unit}{Colors.CLEAR} / {int(total)} {unit}'
        )



"""
Distributions enumeration module.
Operating Systems detection logic.
Interface to `os-release` (through `distro` module).
"""

import os
import platform
import sys

from enum import Enum
from functools import lru_cache
from typing import List, Optional

import distro


class Distributions(Enum):
    """
    This enumeration lists supported operating systems (keys).
    Values contain their respective `distro` identifier.
    See <https://distro.readthedocs.io/en/latest/#distro.id>.
    """
    ALPINE = 'alpine'
    ANDROID = 'android'
    ARCH = 'arch'
    BUNSENLABS = 'bunsenlabs'
    CENTOS = 'centos'
    CRUNCHBANG = 'crunchbang'
    DARWIN = 'darwin'
    DEBIAN = 'debian'
    DEVUAN = 'devuan'
    ELEMENTARY = 'elementary'
    FEDORA = 'fedora'
    FREEBSD = 'freebsd'
    GENTOO = 'gentoo'
    KALI = 'kali'
    MANJARO = 'manjaro'
    NETBSD = 'netbsd'
    NIXOS = 'nixos'
    LINUX = 'linux'
    LINUXMINT = 'linuxmint'
    OPENBSD = 'openbsd'
    OPENSUSE = 'opensuse'
    POP = 'pop'
    PARABOLA = 'parabola'
    RASPBIAN = 'raspbian'
    ROCKY = 'rocky'
    RHEL = 'rhel'
    SLACKWARE = 'slackware'
    UBUNTU = 'ubuntu'
    WINDOWS = 'windows'


    @staticmethod
    def get_identifiers() -> List[str]:
        """Simple getter returning current supported distributions identifiers"""
        return [d.value for d in Distributions.__members__.values()]

    @staticmethod
    @lru_cache(maxsize=None)  # Python < 3.9, `functools.cache` is not yet available.
    def get_local() -> 'Distributions':
        """Entry point of Archey distribution detection logic"""
        distribution = Distributions._vendor_detection()

        # In case nothing got detected the "regular" way...
        if not distribution:
            # Are we running on Darwin (somehow not previously detected by `distro`) ?
            if platform.system() == 'Darwin':
                return Distributions.DARWIN

            # Android systems are currently not being handled by `distro`.
            # At first, we imitate the Python standard library, by checking whether CPython
            #  has been built for Android.
            #  See <https://github.com/python/cpython/search?l=Python&q=getandroidapilevel>
            # As a fallback, we mimic Neofetch behavior, by relying on the file-system.
            #  See <https://github.com/nir0s/distro/issues/253>
            if hasattr(sys, 'getandroidapilevel') \
                or (os.path.isdir('/system/app') and os.path.isdir('/system/priv-app')):
                return Distributions.ANDROID

            # If nothing of the above matched, fall-back on the Linux logo.
            return Distributions.LINUX

        # Below are brain-dead cases for distributions not properly handled by `distro`.
        # One _may_ want to add its own logic to add support for such undetectable systems.
        if distribution == Distributions.DEBIAN:
            # CrunchBang is tagged as _regular_ Debian by `distro`.
            # Below conditions are here to work-around this issue.
            # First condition : CrunchBang-Linux and CrunchBang-Monara.
            # Second condition : CrunchBang++ (CBPP).
            if os.path.isfile('/etc/lsb-release-crunchbang') \
                or os.path.isfile('/usr/bin/cbpp-exit'):
                return Distributions.CRUNCHBANG

        elif distribution == Distributions.UBUNTU:
            # Older Pop!_OS releases (< 20.*) didn't ship their own `ID` (from `os-release`).
            # Thus, they are detected as "regular" Ubuntu distributions.
            # We may here rely on their `NAME` (from `os-release`), which is sufficient.
            if Distributions.get_distro_name(pretty=False) == 'Pop!_OS':
                return Distributions.POP

        return distribution

    @staticmethod
    def _vendor_detection() -> Optional['Distributions']:
        """Main distribution detection logic, relying on `distro`, handling _common_ cases"""
        # Are we running on Windows ?
        if platform.system() == 'Windows':
            return Distributions.WINDOWS

        # Is it a Windows Sub-system Linux (WSL) distribution ?
        # If so, kernel release identifier should keep a trace of it.
        if 'microsoft' in platform.release().lower():
            return Distributions.WINDOWS

        # Is `ID` (from `os-release`) well-known and supported ?
        try:
            return Distributions(distro.id())
        except ValueError:
            pass

        # Is any of `ID_LIKE` (from `os-release`) well-known and supported ?
        # See <https://www.freedesktop.org/software/systemd/man/os-release.html#ID_LIKE=>.
        for id_like in distro.like().split(' '):
            try:
                return Distributions(id_like)
            except ValueError:
                pass

        # Nothing of the above matched, let's return `None` and let the caller handle it.
        return None

    @staticmethod
    def get_distro_name(pretty: bool = True) -> Optional[str]:
        """Simple wrapper to `distro` to return the current distribution _pretty_ name"""
        return distro.name(pretty=pretty) or None

    @staticmethod
    def get_ansi_color() -> Optional[str]:
        """
        Simple wrapper to `distro` to return the distribution preferred ANSI color.
        See <https://www.freedesktop.org/software/systemd/man/os-release.html#ANSI_COLOR=>.
        """
        return distro.os_release_attr('ansi_color') or None



"""Test module for Archey's session user name detection module"""

import unittest
from unittest.mock import MagicMock, patch

from archey.configuration import DEFAULT_CONFIG
from archey.entries.user import User
from archey.test.entries import HelperMethods


class TestUserEntry(unittest.TestCase):
    """
    For this entry, we'll simply mock `getpass.getuser` call.
    If internals happen to fail, `ImportError` might be raised.
    """
    @patch(
        'archey.entries.user.getpass.getuser',
        side_effect=[
            'USERNAME',
            ImportError('pwd', "Sure, you got a good reason...")
        ]
    )
    def test_getenv(self, _):
        """Simple mock, simple test"""
        self.assertEqual(User().value, 'USERNAME')
        self.assertIsNone(User().value)

    @HelperMethods.patch_clean_configuration
    def test_output(self):
        """Simple test for `output` base method"""
        user_instance_mock = HelperMethods.entry_mock(User)

        output_mock = MagicMock()
        User.output(user_instance_mock, output_mock)

        self.assertEqual(
            output_mock.append.call_args[0][1],
            DEFAULT_CONFIG['default_strings']['not_detected']
        )


if __name__ == '__main__':
    unittest.main()



"""CPU information detection class"""

import json
import platform
import re

from subprocess import CalledProcessError, DEVNULL, check_output
from typing import Dict, List

from archey.entry import Entry


class CPU(Entry):
    """
    Parse `/proc/cpuinfo` file to retrieve model names.
    If no information could be retrieved, call `lscpu`.

    `value` attribute is populated as a `list` of `dict`.
    Each `dict` **SHOULD** contain only one entry (CPU model name as key and cores count as value).
    """
    _MODEL_NAME_REGEXP = re.compile(
        r'^model name\s*:\s*(.*)$',
        flags=re.IGNORECASE | re.MULTILINE
    )
    _PHYSICAL_ID_REGEXP = re.compile(
        r'^physical id\s*:\s*(\d+)$',
        flags=re.IGNORECASE | re.MULTILINE
    )
    _THREADS_PER_CORE_REGEXP = re.compile(
        r'^Thread\(s\) per core\s*:\s*(\d+)$',
        flags=re.IGNORECASE | re.MULTILINE
    )
    _CORES_PER_SOCKET_REGEXP = re.compile(
        r'^Core\(s\) per socket\s*:\s*(\d+)$',
        flags=re.IGNORECASE | re.MULTILINE
    )
    _SOCKETS_REGEXP = re.compile(
        r'^Socket\(s\)\s*:\s*(\d+)$',
        flags=re.IGNORECASE | re.MULTILINE
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        if platform.system() == 'Linux':
            self.value = self._parse_proc_cpuinfo()
        else:
            # Darwin or any other BSD-based system.
            self.value = self._parse_system_profiler() or \
                self._parse_sysctl_machdep()

        if not self.value:
            # This test case has been built for some ARM architectures (see #29).
            # Sometimes, `model name` info is not present within `/proc/cpuinfo`.
            # We use the output of `lscpu` program (util-linux-ng) to retrieve it.
            self.value = self._parse_lscpu_output()

    @classmethod
    def _parse_proc_cpuinfo(cls) -> List[Dict[str, int]]:
        """Read `/proc/cpuinfo` and search for CPU model names occurrences"""
        try:
            with open('/proc/cpuinfo', encoding='ASCII') as f_cpu_info:
                cpu_info = f_cpu_info.read()
        except OSError:
            return []

        model_names = cls._MODEL_NAME_REGEXP.findall(cpu_info)
        physical_ids = cls._PHYSICAL_ID_REGEXP.findall(cpu_info)

        cpus_list: List[Dict[str, int]] = []

        # Manually de-duplicates CPUs count.
        for model_name, physical_id in zip(model_names, physical_ids):
            # Sometimes CPU model names contain extra ugly white-spaces.
            model_name = re.sub(r'\s+', ' ', model_name)

            try:
                cpus_list[int(physical_id)][model_name] += 1
            except KeyError:
                # Different CPUs with same physical ids ? Shouldn't happen.
                cpus_list[int(physical_id)][model_name] = 1
            except IndexError:
                cpus_list.append({model_name: 1})

        return cpus_list

    @classmethod
    def _parse_lscpu_output(cls) -> List[Dict[str, int]]:
        """Same operation but from `lscpu` output"""
        try:
            cpu_info = check_output(
                'lscpu',
                env={'LANG': 'C'},
                universal_newlines=True
            )
        except FileNotFoundError:
            return []

        nb_threads = cls._THREADS_PER_CORE_REGEXP.findall(cpu_info)
        nb_cores = cls._CORES_PER_SOCKET_REGEXP.findall(cpu_info)
        nb_sockets = cls._SOCKETS_REGEXP.findall(cpu_info)
        model_names = cls._MODEL_NAME_REGEXP.findall(cpu_info)

        cpus_list = []

        for threads, cores, sockets, model_name in zip(
                nb_threads, nb_cores, nb_sockets, model_names
            ):
            for _ in range(int(sockets)):
                # Sometimes CPU model names contain extra ugly white-spaces.
                cpus_list.append(
                    {re.sub(r'\s+', ' ', model_name): int(threads) * int(cores)}
                )

        return cpus_list

    @staticmethod
    def _parse_system_profiler() -> List[Dict[str, int]]:
        # Parse JSON output from Darwin's `system_profiler` binary.
        try:
            profiler_output = check_output(
                ['system_profiler', '-json', 'SPHardwareDataType'],
                stderr=DEVNULL, universal_newlines=True
            )
        except (FileNotFoundError, CalledProcessError):
            # `-json` is not available before Catalina.
            return []

        cpus_list = []

        for hw_overview in json.loads(profiler_output).get('SPHardwareDataType', []):
            model_name = hw_overview.get('cpu_type')
            nb_cores = hw_overview.get('number_processors')
            nb_sockets = hw_overview.get('packages')
            if not model_name or not nb_cores or not nb_sockets:
                continue

            # Add processor speed (if available).
            proc_speed = hw_overview.get('current_processor_speed')
            if proc_speed:
                model_name += f" @ {proc_speed.replace(',', '.')}"

            # Intel hyper-threading.
            if hw_overview.get('platform_cpu_htt') == 'htt_enabled':
                nb_cores *= 2

            for _ in range(nb_sockets):
                cpus_list.append({model_name: nb_cores})

        return cpus_list

    @staticmethod
    def _parse_sysctl_machdep() -> List[Dict[str, int]]:
        # Runs `sysctl` to fetch some `machdep.cpu.*` keys.
        try:
            sysctl_output = check_output(
                [
                    'sysctl', '-n',
                    'machdep.cpu.brand_string', 'machdep.cpu.core_count'
                ],
                stderr=DEVNULL, universal_newlines=True
            )
        except (FileNotFoundError, CalledProcessError):
            return []

        # `sysctl_output` should exactly contains two lines.
        model_name, nb_cores = sysctl_output.splitlines()
        return [{model_name: int(nb_cores)}]


    def output(self, output):
        """Writes CPUs to `output` based on preferences"""
        # No CPU could be detected.
        if not self.value:
            output.append(self.name, self._default_strings.get('not_detected'))
            return

        entries = []
        for cpus in self.value:
            for model_name, cpu_count in cpus.items():
                if cpu_count > 1 and self.options.get('show_cores', True):
                    entries.append(f'{cpu_count} x {model_name}')
                else:
                    entries.append(model_name)

        if self.options.get('one_line'):
            # One-line output is enabled : Join the results !
            output.append(self.name, ', '.join(entries))
        else:
            # One-line output has been disabled, add one entry per item.
            for entry in entries:
                output.append(self.name, entry)



"""Rocky Linux logo"""

from archey.colors import Colors


COLORS = [Colors.GREEN_BRIGHT]

LOGO = [
    """{c[0]}           __wgliliiligw_,          """,
    """{c[0]}        _williiiiiiliilililw,       """,
    """{c[0]}      _%%iiiiiilililiiiiiiiiiii_    """,
    """{c[0]}    .Qliiiililiiiiiiililililiilm.   """,
    """{c[0]}   _iiiiiliiiiiililiiiiiiiiiiliil,  """,
    """{c[0]}  .lililiiilililiiiilililililiiiii, """,
    """{c[0]} _liiiiiiliiiiiiiliiiiiF{{iiiiiilili,""",
    """{c[0]} jliililiiilililiiili@`  ~ililiiiiiL""",
    """{c[0]} iiiliiiiliiiiiiili>`      ~liililii""",
    """{c[0]} liliiiliiilililii`         -9liiiil""",
    """{c[0]} iiiiiliiliiiiii~             "4lili""",
    """{c[0]} 4ililiiiiilil~|      -w,       )4lf""",
    """{c[0]} -liiiiililiF'       _liig,       )'""",
    """{c[0]}  )iiiliii@`       _QIililig,       """,
    """{c[0]}   )iiii>`       .Qliliiiililw      """,
    """{c[0]}    )<>~       .mliiiiiliiiiiil,    """,
    """{c[0]}             _gllilililiililii~     """,
    """{c[0]}            giliiiiiiiiiiiiT`       """,
    """{c[0]}           -^~$ililili@~~'          """
]



"""Host-name detection class"""

import platform

from typing import Optional

from archey.entry import Entry


class Hostname(Entry):
    """Read system file with fallback on `platform` module to retrieve the system host-name"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.value = self._read_etc_hostname()
        if not self.value:
            self.value = platform.node()

    @staticmethod
    def _read_etc_hostname() -> Optional[str]:
        try:
            with open('/etc/hostname', encoding='UTF-8') as f_hostname:
                return f_hostname.read().rstrip()
        except FileNotFoundError:
            return None
